# -*- coding: utf-8 -*-
"""Numerical/mathemetical related functions.

.. versionadded:: 2.1.0
"""

from __future__ import absolute_import, division

import math
import operator

import pydash as pyd
from .helpers import NoValue, iterator_with_default, iteriteratee, iterator
from ._compat import _range


__all__ = (
    'add',
    'ceil',
    'clamp',
    'divide',
    'floor',
    'max_',
    'max_by',
    'mean',
    'mean_by',
    'median',
    'min_',
    'min_by',
    'moving_mean',
    'multiply',
    'power',
    'round_',
    'scale',
    'slope',
    'std_deviation',
    'sum_',
    'sum_by',
    'subtract',
    'transpose',
    'variance',
    'zscore',
)


INFINITY = float('inf')


def add(a, b):
    """Adds two numbers.

    Args:
        a (number): First number to add.
        b (number): Second number to add.

    Returns:
        number

    Example:

        >>> add(10, 5)
        15

    .. versionadded:: 2.1.0

    .. versionchanged:: 3.3.0
        Support adding two numbers when passed as positional arguments.

    .. versionchanged:: 4.0.0
        Only support two argument addition.
    """
    return a + b


def sum_(collection):
    """Sum each element in `collection`.

    Args:
        collection (list|dict|number): Collection to process or first number to
            add.

    Returns:
        number: Result of summation.

    Example:

        >>> sum_([1, 2, 3, 4])
        10

    .. versionadded:: 2.1.0

    .. versionchanged:: 3.3.0
        Support adding two numbers when passed as positional arguments.

    .. versionchanged:: 4.0.0
        Move iteratee support to :func:`sum_by`. Move two argument addition to
        :func:`add`.
    """
    return sum_by(collection)


def sum_by(collection, iteratee=None):
    """Sum each element in `collection`. If iteratee is passed, each element of
    `collection` is passed through a iteratee before the summation is computed.

    Args:
        collection (list|dict|number): Collection to process or first number to
            add.
        iteratee (mixed|number, optional): Iteratee applied per iteration or
            second number to add.

    Returns:
        number: Result of summation.

    Example:

        >>> sum_by([1, 2, 3, 4], lambda x: x ** 2)
        30

    .. versionadded:: 4.0.0
    """
    return sum(result[0] for result in iteriteratee(collection, iteratee))


def mean(collection):
    """Calculate arithmetic mean of each element in `collection`.

    Args:
        collection (list|dict): Collection to process.

    Returns:
        float: Result of mean.

    Example:

        >>> mean([1, 2, 3, 4])
        2.5

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0

        - Removed ``average`` and ``avg`` aliases.
        - Moved iteratee functionality to :func:`mean_by`.
    """
    return mean_by(collection)


def mean_by(collection, iteratee=None):
    """Calculate arithmetic mean of each element in `collection`. If iteratee
    is passed, each element of `collection` is passed through a iteratee before
    the mean is computed.

    Args:
        collection (list|dict): Collection to process.
        iteratee (mixed, optional): Iteratee applied per iteration.

    Returns:
        float: Result of mean.

    Example:

        >>> mean_by([1, 2, 3, 4], lambda x: x ** 2)
        7.5

    .. versionadded:: 4.0.0
    """
    return sum_by(collection, iteratee) / pyd.size(collection)


def ceil(x, precision=0):
    """Round number up to precision.

    Args:
        x (number): Number to round up.
        precision (int, optional): Rounding precision. Defaults to ``0``.

    Returns:
        int: Number rounded up.

    Example:

        >>> ceil(3.275) == 4.0
        True
        >>> ceil(3.215, 1) == 3.3
        True
        >>> ceil(6.004, 2) == 6.01
        True

    .. versionadded:: 3.3.0
    """
    return rounder(math.ceil, x, precision)


def clamp(x, lower, upper=None):
    """Clamps number within the inclusive lower and upper bounds.

    Args:
        x (number): Number to clamp.
        lower (number, optional): Lower bound.
        upper (number): Upper bound

    Returns:
        number

    Example:

        >>> clamp(-10, -5, 5)
        -5
        >>> clamp(10, -5, 5)
        5
        >>> clamp(10, 5)
        5
        >>> clamp(-10, 5)
        -10

    .. versionadded:: 4.0.0
    """
    if upper is None:
        upper = lower
        lower = x

    if x < lower:
        x = lower
    elif x > upper:
        x = upper

    return x


def divide(dividend, divisor):
    """Divide two numbers.

    Args:
        dividend (int/float): The first number in a division.
        divisor (int/float): The second number in a division.

    Returns:
        int/float: Returns the quotient.

    Example:

        >>> divide(20, 5)
        4.0
        >>> divide(1.5, 3)
        0.5
        >>> divide(None, None)
        1.0
        >>> divide(5, None)
        5.0

    .. versionadded:: 4.0.0
    """
    return call_math_operator(dividend, divisor, operator.truediv, 1)


def floor(x, precision=0):
    """Round number down to precision.

    Args:
        x (number): Number to round down.
        precision (int, optional): Rounding precision. Defaults to ``0``.

    Returns:
        int: Number rounded down.

    Example:

        >>> floor(3.75) == 3.0
        True
        >>> floor(3.215, 1) == 3.2
        True
        >>> floor(0.046, 2) == 0.04
        True

    .. versionadded:: 3.3.0
    """
    return rounder(math.floor, x, precision)


def max_(collection, default=NoValue):
    """Retrieves the maximum value of a `collection`.

    Args:
        collection (list|dict): Collection to iterate over.
        default (mixed, optional): Value to return if `collection` is empty.

    Returns:
        mixed: Maximum value.

    Example:

        >>> max_([1, 2, 3, 4])
        4
        >>> max_([], default=-1)
        -1

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee iteratee support to :func:`max_by`.
    """
    return max_by(collection, default=default)


def max_by(collection, iteratee=None, default=NoValue):
    """Retrieves the maximum value of a `collection`.

    Args:
        collection (list|dict): Collection to iterate over.
        iteratee (mixed, optional): Iteratee applied per iteration.
        default (mixed, optional): Value to return if `collection` is empty.

    Returns:
        mixed: Maximum value.

    Example:

        >>> max_by([1.0, 1.5, 1.8], math.floor)
        1.0
        >>> max_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a')
        {'a': 3}
        >>> max_by([], default=-1)
        -1

    .. versionadded:: 4.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()

    return max(iterator_with_default(collection, default),
               key=pyd.iteratee(iteratee))


def median(collection, iteratee=None):
    """Calculate median of each element in `collection`. If iteratee is passed,
    each element of `collection` is passed through a iteratee before the
    median is computed.

    Args:
        collection (list|dict): Collection to process.
        iteratee (mixed, optional): Iteratee applied per iteration.

    Returns:
        float: Result of median.

    Example:

        >>> median([1, 2, 3, 4, 5])
        3
        >>> median([1, 2, 3, 4])
        2.5

    .. versionadded:: 2.1.0
    """
    length = len(collection)
    middle = (length + 1) / 2
    collection = [ret[0] for ret in iteriteratee(sorted(collection), iteratee)]

    if pyd.is_odd(length):
        result = collection[int(middle - 1)]
    else:
        left = int(middle - 1.5)
        right = int(middle - 0.5)
        result = (collection[left] + collection[right]) / 2

    return result


def min_(collection, default=NoValue):
    """Retrieves the minimum value of a `collection`.

    Args:
        collection (list|dict): Collection to iterate over.
        default (mixed, optional): Value to return if `collection` is empty.

    Returns:
        mixed: Minimum value.

    Example:

        >>> min_([1, 2, 3, 4])
        1
        >>> min_([], default=100)
        100

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Moved iteratee iteratee support to :func:`min_by`.
    """
    return min_by(collection, default=default)


def min_by(collection, iteratee=None, default=NoValue):
    """Retrieves the minimum value of a `collection`.

    Args:
        collection (list|dict): Collection to iterate over.
        iteratee (mixed, optional): Iteratee applied per iteration.
        default (mixed, optional): Value to return if `collection` is empty.

    Returns:
        mixed: Minimum value.

    Example:

        >>> min_by([1.8, 1.5, 1.0], math.floor)
        1.8
        >>> min_by([{'a': 1}, {'a': 2}, {'a': 3}], 'a')
        {'a': 1}
        >>> min_by([], default=100)
        100

    .. versionadded:: 4.0.0
    """
    if isinstance(collection, dict):
        collection = collection.values()
    return min(iterator_with_default(collection, default),
               key=pyd.iteratee(iteratee))


def moving_mean(array, size):
    """Calculate moving mean of each element of `array`.

    Args:
        array (list): List to process.
        size (int): Window size.

    Returns:
        list: Result of moving average.

    Example:

        >>> moving_mean(range(10), 1)
        [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        >>> moving_mean(range(10), 5)
        [2.0, 3.0, 4.0, 5.0, 6.0, 7.0]
        >>> moving_mean(range(10), 10)
        [4.5]

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Rename to ``moving_mean`` and remove ``moving_average`` and
        ``moving_avg`` aliases.
    """
    result = []
    size = int(size)

    for i in _range(size - 1, len(array) + 1):
        window = array[i - size:i]

        if len(window) == size:
            result.append(mean(window))

    return result


def multiply(multiplier, multiplicand):
    """Multiply two numbers.

    Args:
        multiplier (int/float): The first number in a multiplication.
        multiplicand (int/float): The second number in a multiplication.

    Returns:
        int/float: Returns the product.

    Example:

        >>> multiply(4, 5)
        20
        >>> multiply(10, 4)
        40
        >>> multiply(None, 10)
        10
        >>> multiply(None, None)
        1

    .. versionadded:: 4.0.0
    """
    return call_math_operator(multiplier, multiplicand, operator.mul, 1)


def power(x, n):
    """Calculate exponentiation of `x` raised to the `n` power.

    Args:
        x (number): Base number.
        n (number): Exponent.

    Returns:
        number: Result of calculation.

    Example:

        >>> power(5, 2)
        25
        >>> power(12.5, 3)
        1953.125

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Removed alias ``pow_``.
    """
    if pyd.is_number(x):
        result = pow(x, n)
    elif pyd.is_list(x):
        result = pyd.map_(x, lambda item: pow(item, n))
    else:
        result = None

    return result


def round_(x, precision=0):
    """Round number to precision.

    Args:
        x (number): Number to round.
        precision (int, optional): Rounding precision. Defaults to ``0``.

    Returns:
        int: Rounded number.

    Example:

        >>> round_(3.275) == 3.0
        True
        >>> round_(3.275, 1) == 3.3
        True

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Remove alias ``curve``.
    """
    return rounder(round, x, precision)


def scale(array, maximum=1):
    """Scale list of value to a maximum number.

    Args:
        array (list): Numbers to scale.
        maximum (number): Maximum scale value.

    Returns:
        list: Scaled numbers.

    Example:

        >>> scale([1, 2, 3, 4])
        [0.25, 0.5, 0.75, 1.0]
        >>> scale([1, 2, 3, 4], 1)
        [0.25, 0.5, 0.75, 1.0]
        >>> scale([1, 2, 3, 4], 4)
        [1.0, 2.0, 3.0, 4.0]
        >>> scale([1, 2, 3, 4], 2)
        [0.5, 1.0, 1.5, 2.0]

    .. versionadded:: 2.1.0
    """
    array_max = max(array)
    return pyd.map_(array, lambda item: item * (maximum / array_max))


def slope(point1, point2):
    """Calculate the slope between two points.

    Args:
        point1 (list|tuple): X and Y coordinates of first point.
        point2 (list|tuple): X and Y cooredinates of second point.

    Returns:
        float: Calculated slope.

    Example:

        >>> slope((1, 2), (4, 8))
        2.0

    .. versionadded:: 2.1.0
    """
    x1, y1 = point1[0], point1[1]
    x2, y2 = point2[0], point2[1]

    if x1 == x2:
        result = INFINITY
    else:
        result = (y2 - y1) / (x2 - x1)

    return result


def std_deviation(array):
    """Calculate standard deviation of list of numbers.

    Args:
        array (list): List to process.

    Returns:
        float: Calculated standard deviation.

    Example:

        >>> round(std_deviation([1, 18, 20, 4]), 2) == 8.35
        True

    .. versionadded:: 2.1.0

    .. versionchanged:: 4.0.0
        Remove alias ``sigma``.
    """
    return math.sqrt(variance(array))


def subtract(minuend, subtrahend):
    """Subtracts two numbers.

    Args:
        minuend (int/float): Value passed in by the user.
        subtrahend (int/float): Value passed in by the user.

    Returns:
        int/float: Result of the difference from the given values.

    Example:

        >>> subtract(10, 5)
        5
        >>> subtract(-10, 4)
        -14
        >>> subtract(2, 0.5)
        1.5

    .. versionadded:: 4.0.0
    """
    return call_math_operator(minuend, subtrahend, operator.sub, 0)


def transpose(array):
    """Transpose the elements of `array`.

    Args:
        array (list): List to process.

    Returns:
        list: Transposed list.

    Example:

        >>> transpose([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        [[1, 4, 7], [2, 5, 8], [3, 6, 9]]

    .. versionadded:: 2.1.0
    """
    trans = []

    for y, row in iterator(array):
        for x, col in iterator(row):
            trans = pyd.set_(trans, [x, y], col)

    return trans


def variance(array):
    """Calculate the variance of the elements in `array`.

    Args:
        array (list): List to process.

    Returns:
        float: Calculated variance.

    Example:

        >>> variance([1, 18, 20, 4])
        69.6875

    .. versionadded:: 2.1.0
    """
    avg = mean(array)

    def var(x):
        return power(x - avg, 2)

    return pyd._(array).map_(var).mean().value()


def zscore(collection, iteratee=None):
    """Calculate the standard score assuming normal distribution. If iteratee
    is passed, each element of `collection` is passed through a iteratee before
    the standard score is computed.

    Args:
        collection (list|dict): Collection to process.
        iteratee (mixed, optional): Iteratee applied per iteration.

    Returns:
        float: Calculated standard score.

    Example:

        >>> results = zscore([1, 2, 3])

        # [-1.224744871391589, 0.0, 1.224744871391589]

    .. versionadded:: 2.1.0
    """
    array = pyd.map_(collection, iteratee)
    avg = mean(array)
    sig = std_deviation(array)

    return pyd.map_(array, lambda item: (item - avg) / sig)


#
# Utility methods not a part of the main API
#

def call_math_operator(value1, value2, op, default):
    """Return the result of the math operation on the given values."""
    if not value1:
        value1 = default

    if not value2:
        value2 = default

    if not pyd.is_number(value1):
        try:
            value1 = float(value1)
        except Exception:
            pass

    if not pyd.is_number(value2):
        try:
            value2 = float(value2)
        except Exception:
            pass

    return op(value1, value2)


def rounder(func, x, precision):
    precision = pow(10, precision)

    def rounder_func(item):
        return func(item * precision) / precision

    result = None

    if pyd.is_number(x):
        result = rounder_func(x)
    elif pyd.is_iterable(x):
        try:
            result = pyd.map_(x, rounder_func)
        except TypeError:
            pass

    return result



# -*- coding: utf-8 -*-
"""Functions that operate on lists.

.. versionadded:: 1.0.0
"""

from __future__ import absolute_import

from bisect import bisect_left, bisect_right
from math import ceil

import pydash as pyd
from .helpers import parse_iteratee, iteriteratee, base_get
from ._compat import cmp_to_key, string_types


__all__ = (
    'chunk',
    'compact',
    'concat',
    'difference',
    'difference_by',
    'difference_with',
    'drop',
    'drop_right',
    'drop_right_while',
    'drop_while',
    'duplicates',
    'fill',
    'find_index',
    'find_last_index',
    'flatten',
    'flatten_deep',
    'flatten_depth',
    'from_pairs',
    'head',
    'index_of',
    'initial',
    'intercalate',
    'interleave',
    'intersection',
    'intersection_by',
    'intersection_with',
    'intersperse',
    'last',
    'last_index_of',
    'mapcat',
    'nth',
    'pull',
    'pull_all',
    'pull_all_by',
    'pull_all_with',
    'pull_at',
    'push',
    'remove',
    'reverse',
    'shift',
    'slice_',
    'sort',
    'sorted_index',
    'sorted_index_by',
    'sorted_index_of',
    'sorted_last_index',
    'sorted_last_index_by',
    'sorted_last_index_of',
    'sorted_uniq',
    'sorted_uniq_by',
    'splice',
    'split_at',
    'tail',
    'take',
    'take_right',
    'take_right_while',
    'take_while',
    'union',
    'union_by',
    'union_with',
    'uniq',
    'uniq_by',
    'uniq_with',
    'unshift',
    'unzip',
    'unzip_with',
    'without',
    'xor',
    'xor_by',
    'xor_with',
    'zip_',
    'zip_object',
    'zip_object_deep',
    'zip_with'
)


def chunk(array, size=1):
    """Creates a list of elements split into groups the length of `size`. If
    `array` can't be split evenly, the final chunk will be the remaining
    elements.

    Args:
        array (list): List to chunk.
        size (int, optional): Chunk size. Defaults to ``1``.

    Returns:
        list: New list containing chunks of `array`.

    Example:

        >>> chunk([1, 2, 3, 4, 5], 2)
        [[1, 2], [3, 4], [5]]

    .. versionadded:: 1.1.0
    """
    chunks = int(ceil(len(array) / float(size)))
    return [array[i * size:(i + 1) * size] for i in range(chunks)]


def compact(array):
    """Creates a list with all falsey values of array removed.

    Args:
        array (list): List to compact.

    Returns:
        list: Compacted list.

    Example:

        >>> compact(['', 1, 0, True, False, None])
        [1, True]

    .. versionadded:: 1.0.0
    """
    return [item for item in array if item]


def concat(*arrays):
    """Concatenates zero or more lists into one.

    Args:
        arrays (list): Lists to concatenate.

    Returns:
        list: Concatenated list.

    Example:

        >>> concat([1, 2], [3, 4], [[5], [6]])
        [1, 2, 3, 4, [5], [6]]

    .. versionadded:: 2.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``cat`` to ``concat``.
    """
    return flatten(arrays)


def difference(array, *others):
    """Creates a list of list elements not present in others.

    Args:
        array (list): List to process.
        others (list): Lists to check.

    Returns:
        list: Difference between `others`.

    Example:

        >>> difference([1, 2, 3], [1], [2])
        [3]

    .. versionadded:: 1.0.0
    """
    return difference_with(array, *others)


def difference_by(array, *others, **kargs):
    """This method is like :func:`difference` except that it accepts an
    iteratee which is invoked for each element of each array to generate the
    criterion by which they're compared. The order and references of result
    values are determined by `array`. The iteratee is invoked with one
    argument: ``(value)``.

    Args:
        array (list): The array to find the difference of.
        others (list): Lists to check for difference with `array`.

    Keyword Args:
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: Difference between `others`.

    Example:

        >>> difference_by([1.2, 1.5, 1.7, 2.8], [0.9, 3.2], round)
        [1.5, 1.7]

    .. versionadded:: 4.0.0
    """
    array = array[:]

    if not others:
        return array

    # Check if last other is a potential iteratee.
    iteratee, others = parse_iteratee('iteratee', *others, **kargs)

    for other in others:
        if not other:
            continue
        array = list(iterdifference(array, other, iteratee=iteratee))

    return array


def difference_with(array, *others, **kargs):
    """This method is like :func:`difference` except that it accepts a
    comparator which is invoked to compare the elements of all arrays. The
    order and references of result values are determined by the first array.
    The comparator is invoked with two arguments: ``(arr_val, oth_val)``.

    Args:
        array (list): The array to find the difference of.
        others (list): Lists to check for difference with `array`.

    Keyword Args:
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: Difference between `others`.

    Example:

        >>> array = ['apple', 'banana', 'pear']
        >>> others = (['avocado', 'pumpkin'], ['peach'])
        >>> comparator = lambda a, b: a[0] == b[0]
        >>> difference_with(array, *others, comparator=comparator)
        ['banana']

    .. versionadded:: 4.0.0
    """
    array = array[:]

    if not others:
        return array

    comparator = kargs.get('comparator')
    last_other = others[-1]

    # Check if last other is a comparator.
    if comparator is None and (callable(last_other) or last_other is None):
        comparator = last_other
        others = others[:-1]

    for other in others:
        if not other:
            continue
        array = list(iterdifference(array, other, comparator=comparator))

    return array


def drop(array, n=1):
    """Creates a slice of `array` with `n` elements dropped from the beginning.

    Args:
        array (list): List to process.
        n (int, optional): Number of elements to drop. Defaults to ``1``.

    Returns:
        list: Dropped list.

    Example:

        >>> drop([1, 2, 3, 4], 2)
        [3, 4]

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Added ``n`` argument and removed as alias of :func:`rest`.

    .. versionchanged:: 3.0.0
        Made ``n`` default to ``1``.
    """
    return drop_while(array, lambda _, index: index < n)


def drop_right(array, n=1):
    """Creates a slice of `array` with `n` elements dropped from the end.

    Args:
        array (list): List to process.
        n (int, optional): Number of elements to drop. Defaults to ``1``.

    Returns:
        list: Dropped list.

    Example:

        >>> drop_right([1, 2, 3, 4], 2)
        [1, 2]

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Made ``n`` default to ``1``.
    """
    length = len(array)
    return drop_right_while(array, lambda _, index: (length - index) <= n)


def drop_right_while(array, predicate=None):
    """Creates a slice of `array` excluding elements dropped from the end.
    Elements are dropped until the `predicate` returns falsey. The `predicate`
    is invoked with three arguments: ``(value, index, array)``.

    Args:
        array (list): List to process.
        predicate (mixed): Predicate called per iteration

    Returns:
        list: Dropped list.

    Example:

        >>> drop_right_while([1, 2, 3, 4], lambda x: x >= 3)
        [1, 2]

    .. versionadded:: 1.1.0
    """
    n = len(array)
    for is_true, _, _, _ in iteriteratee(array, predicate, reverse=True):
        if is_true:
            n -= 1
        else:
            break

    return array[:n]


def drop_while(array, predicate=None):
    """Creates a slice of `array` excluding elements dropped from the
    beginning. Elements are dropped until the `predicate` returns falsey. The
    `predicate` is invoked with three arguments: ``(value, index, array)``.

    Args:
        array (list): List to process.
        predicate (mixed): Predicate called per iteration

    Returns:
        list: Dropped list.

    Example:

        >>> drop_while([1, 2, 3, 4], lambda x: x < 3)
        [3, 4]

    .. versionadded:: 1.1.0
    """
    n = 0
    for is_true, _, _, _ in iteriteratee(array, predicate):
        if is_true:
            n += 1
        else:
            break

    return array[n:]


def duplicates(array, iteratee=None):
    """Creates a unique list of duplicate values from `array`. If iteratee is
    passed, each element of array is passed through a iteratee before
    duplicates are computed. The iteratee is invoked with three arguments:
    ``(value, index, array)``. If an object path is passed for iteratee, the
    created iteratee will return the path value of the given element. If an
    object is passed for iteratee, the created filter style iteratee will
    return ``True`` for elements that have the properties of the given object,
    else ``False``.

    Args:
        array (list): List to process.
        iteratee (mixed, optional): Iteratee applied per iteration.

    Returns:
        list: List of duplicates.

    Example:

        >>> duplicates([0, 1, 3, 2, 3, 1])
        [3, 1]

    .. versionadded:: 3.0.0
    """
    if iteratee:
        cbk = pyd.iteratee(iteratee)
        computed = [cbk(item) for item in array]
    else:
        computed = array

    # NOTE: Using array[i] instead of item since iteratee could have modified
    # returned item values.
    lst = uniq([array[i] for i, _ in iterduplicates(computed)])

    return lst


def fill(array, value, start=0, end=None):
    """Fills elements of array with value from `start` up to, but not
    including, `end`.

    Args:
        array (list): List to fill.
        value (mixed): Value to fill with.
        start (int, optional): Index to start filling. Defaults to ``0``.
        end (int, optional): Index to end filling. Defaults to ``len(array)``.

    Returns:
        list: Filled `array`.

    Example:

        >>> fill([1, 2, 3, 4, 5], 0)
        [0, 0, 0, 0, 0]
        >>> fill([1, 2, 3, 4, 5], 0, 1, 3)
        [1, 0, 0, 4, 5]
        >>> fill([1, 2, 3, 4, 5], 0, 0, 100)
        [0, 0, 0, 0, 0]

    Warning:
        `array` is modified in place.

    .. versionadded:: 3.1.0
    """
    if end is None:
        end = len(array)
    else:
        end = min([end, len(array)])

    # Use this style of assignment so that `array` is mutated.
    array[:] = array[:start] + [value] * len(array[start:end]) + array[end:]
    return array


def find_index(array, predicate=None):
    """This method is similar to :func:`pydash.collections.find`, except
    that it returns the index of the element that passes the predicate check,
    instead of the element itself.

    Args:
        array (list): List to process.
        predicate (mixed, optional): Predicate applied per iteration.

    Returns:
        int: Index of found item or ``-1`` if not found.

    Example:

        >>> find_index([1, 2, 3, 4], lambda x: x >= 3)
        2
        >>> find_index([1, 2, 3, 4], lambda x: x > 4)
        -1

    .. versionadded:: 1.0.0
    """
    search = (i for is_true, _, i, _ in iteriteratee(array, predicate)
              if is_true)
    return next(search, -1)


def find_last_index(array, predicate=None):
    """This method is similar to :func:`find_index`, except that it iterates
    over elements from right to left.

    Args:
        array (list): List to process.
        predicate (mixed, optional): Predicate applied per iteration.

    Returns:
        int: Index of found item or ``-1`` if not found.

    Example:

        >>> find_last_index([1, 2, 3, 4], lambda x: x >= 3)
        3
        >>> find_index([1, 2, 3, 4], lambda x: x > 4)
        -1

    .. versionadded:: 1.0.0
    """
    search = (i for is_true, _, i, _ in iteriteratee(array,
                                                     predicate,
                                                     reverse=True)
              if is_true)
    return next(search, -1)


def flatten(array):
    """Flattens a nested array. If `is_deep` is ``True`` the array is
    recursively flattened, otherwise it is only flattened a single level.

    Args:
        array (list): List to flatten.

    Returns:
        list: Flattened list.

    Example:

        >>> flatten([[1], [2, [3]], [[4]]])
        [1, 2, [3], [4]]


    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0
        Removed `callback` option. Added ``is_deep`` option. Made it shallow
        by default.

    .. versionchanged:: 4.0.0
        Removed ``is_deep`` option. Use :func:`flatten_deep` instead.
    """
    return flatten_depth(array, depth=1)


def flatten_deep(array):
    """Flattens a nested array recursively. This is the same as calling
    ``flatten(array, is_deep=True)``.

    Args:
        array (list): List to flatten.

    Returns:
        list: Flattened list.

    Example:

        >>> flatten_deep([[1], [2, [3]], [[4]]])
        [1, 2, 3, 4]

    .. versionadded:: 2.0.0
    """
    return flatten_depth(array, depth=-1)


def flatten_depth(array, depth=1):
    """Recursively flatten `array` up to `depth` times.

    Args:
        array (list): List to flatten.
        depth (int, optional): Depth to flatten to. Defaults to ``1``.

    Returns:
        list: Flattened list.

    Example:

        >>> flatten_depth([[[1], [2, [3]], [[4]]]], 1)
        [[1], [2, [3]], [[4]]]
        >>> flatten_depth([[[1], [2, [3]], [[4]]]], 2)
        [1, 2, [3], [4]]
        >>> flatten_depth([[[1], [2, [3]], [[4]]]], 3)
        [1, 2, 3, 4]
        >>> flatten_depth([[[1], [2, [3]], [[4]]]], 4)
        [1, 2, 3, 4]

    .. versionadded:: 4.0.0
    """
    return list(iterflatten(array, depth=depth))


def from_pairs(pairs):
    """Returns a dict from the given list of pairs.

    Args:
        pairs (list): List of key-value pairs.

    Returns:
        dict

    Example:

        >>> from_pairs([['a', 1], ['b', 2]]) == {'a': 1, 'b': 2}
        True

    .. versionadded:: 4.0.0
    """
    return dict(pairs)


def head(array):
    """Return the first element of `array`.

    Args:
        array (list): List to process.

    Returns:
        mixed: First element of list.

    Example:

        >>> head([1, 2, 3, 4])
        1

    .. versionadded:: 1.0.0

    .. versionchanged::
        Renamed from ``first`` to ``head``.
    """
    return base_get(array, 0, default=None)


def index_of(array, value, from_index=0):
    """Gets the index at which the first occurrence of value is found.

    Args:
        array (list): List to search.
        value (mixed): Value to search for.
        from_index (int, optional): Index to search from.

    Returns:
        int: Index of found item or ``-1`` if not found.

    Example:

        >>> index_of([1, 2, 3, 4], 2)
        1
        >>> index_of([2, 1, 2, 3], 2, from_index=1)
        2

    .. versionadded:: 1.0.0
    """
    try:
        return array.index(value, from_index)
    except ValueError:
        return -1


def initial(array):
    """Return all but the last element of `array`.

    Args:
        array (list): List to process.

    Returns:
        list: Initial part of `array`.

    Example:

        >>> initial([1, 2, 3, 4])
        [1, 2, 3]

    .. versionadded:: 1.0.0
    """
    return array[:-1]


def intercalate(array, separator):
    """Like :func:`intersperse` for lists of lists but shallowly flattening the
    result.

    Args:
        array (list): List to intercalate.
        separator (mixed): Element to insert.

    Returns:
        list: Intercalated list.

    Example:

        >>> intercalate([1, [2], [3], 4], 'x')
        [1, 'x', 2, 'x', 3, 'x', 4]


    .. versionadded:: 2.0.0
    """
    return flatten(intersperse(array, separator))


def interleave(*arrays):
    """Merge multiple lists into a single list by inserting the next element of
    each list by sequential round-robin into the new list.

    Args:
        arrays (list): Lists to interleave.

    Retruns:
        list: Interleaved list.

    Example:

        >>> interleave([1, 2, 3], [4, 5, 6], [7, 8, 9])
        [1, 4, 7, 2, 5, 8, 3, 6, 9]

    .. versionadded:: 2.0.0
    """
    return list(iterinterleave(*arrays))


def intersection(array, *others):
    """Computes the intersection of all the passed-in arrays.

    Args:
        array (list): The array to find the intersection of.
        others (list): Lists to check for intersection with `array`.

    Returns:
        list: Intersection of provided lists.

    Example:

        >>> intersection([1, 2, 3], [1, 2, 3, 4, 5], [2, 3])
        [2, 3]

        >>> intersection([1, 2, 3])
        [1, 2, 3]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Support finding intersection of unhashable types.
    """
    return intersection_with(array, *others)


def intersection_by(array, *others, **kargs):
    """This method is like :func:`intersection` except that it accepts an
    iteratee which is invoked for each element of each array to generate the
    criterion by which they're compared. The order and references of result
    values are determined by `array`. The iteratee is invoked with one
    argument: ``(value)``.

    Args:
        array (list): The array to find the intersection of.
        others (list): Lists to check for intersection with `array`.

    Keyword Args:
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: Intersection of provided lists.

    Example:

        >>> intersection_by([1.2, 1.5, 1.7, 2.8], [0.9, 3.2], round)
        [1.2, 2.8]

    .. versionadded:: 4.0.0
    """
    array = array[:]

    if not others:
        return array

    iteratee, others = parse_iteratee('iteratee', *others, **kargs)

    # Sort by smallest list length to make intersection faster.
    others = sorted(others, key=lambda other: len(other))

    for other in others:
        array = list(iterintersection(array, other, iteratee=iteratee))
        if not array:
            break

    return array


def intersection_with(array, *others, **kargs):
    """This method is like :func:`intersection` except that it accepts a
    comparator which is invoked to compare the elements of all arrays. The
    order and references of result values are determined by the first array.
    The comparator is invoked with two arguments: ``(arr_val, oth_val)``.

    Args:
        array (list): The array to find the intersection of.
        others (list): Lists to check for intersection with `array`.

    Keyword Args:
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: Intersection of provided lists.

    Example:

        >>> array = ['apple', 'banana', 'pear']
        >>> others = (['avocado', 'pumpkin'], ['peach'])
        >>> comparator = lambda a, b: a[0] == b[0]
        >>> intersection_with(array, *others, comparator=comparator)
        ['pear']

    .. versionadded:: 4.0.0
    """
    array = array[:]

    if not others:
        return array

    comparator, others = parse_iteratee('comparator', *others, **kargs)

    # Sort by smallest list length to reduce to intersection faster.
    others = sorted(others, key=lambda other: len(other))

    for other in others:
        array = list(iterintersection(array, other, comparator=comparator))
        if not array:
            break

    return array


def intersperse(array, separator):
    """Insert a separating element between the elements of `array`.

    Args:
        array (list): List to intersperse.
        separator (mixed): Element to insert.

    Returns:
        list: Interspersed list.

    Example:

        >>> intersperse([1, [2], [3], 4], 'x')
        [1, 'x', [2], 'x', [3], 'x', 4]

    .. versionadded:: 2.0.0
    """
    return list(iterintersperse(array, separator))


def last(array):
    """Return the last element of `array`.

    Args:
        array (list): List to process.

    Returns:
        mixed: Last part of `array`.

    Example:

        >>> last([1, 2, 3, 4])
        4

    .. versionadded:: 1.0.0
    """
    return base_get(array, -1, default=None)


def last_index_of(array, value, from_index=None):
    """Gets the index at which the last occurrence of value is found.

    Args:
        array (list): List to search.
        value (mixed): Value to search for.
        from_index (int, optional): Index to search from.

    Returns:
        int: Index of found item or ``False`` if not found.

    Example:

        >>> last_index_of([1, 2, 2, 4], 2)
        2
        >>> last_index_of([1, 2, 2, 4], 2, from_index=1)
        1

    .. versionadded:: 1.0.0
    """
    index = array_len = len(array)

    try:
        from_index = int(from_index)
    except (TypeError, ValueError):
        pass
    else:
        # Set starting index base on from_index offset.
        index = (max(0, index + from_index) if from_index < 0
                 else min(from_index, index - 1))

    while index:
        if index < array_len and array[index] == value:
            return index
        index -= 1
    return -1


def mapcat(array, iteratee=None):
    """Map a iteratee to each element of a list and concatenate the results
    into a single list using :func:`cat`.

    Args:
        array (list): List to map and concatenate.
        iteratee (mixed): Iteratee to apply to each element.

    Returns:
        list: Mapped and concatenated list.

    Example:

        >>> mapcat(range(4), lambda x: list(range(x)))
        [0, 0, 1, 0, 1, 2]

    .. versionadded:: 2.0.0
    """
    return concat(*pyd.map_(array, iteratee))


def nth(array, pos=0):
    """Gets the element at index n of array.

    Args:
        array (list): List passed in by the user.
        pos (int): Index of element to return.

    Returns:
        mixed: Returns the element at :attr:`pos`.

    Example:

        >>> nth([1, 2, 3], 0)
        1
        >>> nth([3, 4, 5, 6], 2)
        5
        >>> nth([11, 22, 33], -1)
        33
        >>> nth([11, 22, 33])
        11

    .. versionadded:: 4.0.0
    """
    return pyd.get(array, pos)


def pop(array, index=-1):
    """Remove element of array at `index` and return element.

    Args:
        array (list): List to pop from.
        index (int, optional): Index to remove element from. Defaults to
            ``-1``.

    Returns:
        mixed: Value at `index`.

    Warning:
        `array` is modified in place.

    Example:

        >>> array = [1, 2, 3, 4]
        >>> item = pop(array)
        >>> item
        4
        >>> array
        [1, 2, 3]
        >>> item = pop(array, index=0)
        >>> item
        1
        >>> array
        [2, 3]

    .. versionadded:: 2.2.0
    """
    return array.pop(index)


def pull(array, *values):
    """Removes all provided values from the given array.

    Args:
        array (list): List to pull from.
        values (mixed): Values to remove.

    Returns:
        list: Modified `array`.

    Warning:
        `array` is modified in place.

    Example:

        >>> pull([1, 2, 2, 3, 3, 4], 2, 3)
        [1, 4]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        :func:`pull` method now calls :func:`pull_all` method for the desired
        functionality.
    """
    return pull_all(array, values)


def pull_all(array, values):
    """Removes all provided values from the given array.

    Args:
        array (list): Array to modify.
        values (list): Values to remove.

    Returns:
        list: Modified `array`.

    Example:

        >>> pull_all([1, 2, 2, 3, 3, 4], [2, 3])
        [1, 4]

    .. versionadded:: 4.0.0
    """
    # Use this style of assignment so that `array` is mutated.
    array[:] = without(array, *values)
    return array


def pull_all_by(array, values, iteratee=None):
    """This method is like :func:`pull_all` except that it accepts iteratee
    which is invoked for each element of array and values to generate the
    criterion by which they're compared. The iteratee is invoked with one
    argument: ``(value)``.

    Args:
        array (list): Array to modify.
        values (list): Values to remove.
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: Modified `array`.

    Example:

        >>> array = [{'x': 1}, {'x': 2}, {'x': 3}, {'x': 1}]
        >>> pull_all_by(array, [{'x': 1}, {'x': 3}], 'x')
        [{'x': 2}]

    .. versionadded:: 4.0.0
    """
    values = difference(array,
                        difference_by(array, values, iteratee=iteratee))
    return pull_all(array, values)


def pull_all_with(array, values, comparator=None):
    """This method is like :func:`pull_all` except that it accepts comparator
    which is invoked to compare elements of array to values. The comparator is
    invoked with two arguments: ``(arr_val, oth_val)``.

    Args:
        array (list): Array to modify.
        values (list): Values to remove.
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: Modified `array`.

    Example:

        >>> array = [{'x': 1, 'y': 2}, {'x': 3, 'y': 4}, {'x': 5, 'y': 6}]
        >>> res = pull_all_with(array, [{'x': 3, 'y': 4}], lambda a, b: a == b)
        >>> res == [{'x': 1, 'y': 2}, {'x': 5, 'y': 6}]
        True
        >>> array = [{'x': 1, 'y': 2}, {'x': 3, 'y': 4}, {'x': 5, 'y': 6}]
        >>> res = pull_all_with(array, [{'x': 3, 'y': 4}], lambda a, b: a != b)
        >>> res == [{'x': 3, 'y': 4}]
        True

    .. versionadded:: 4.0.0
    """
    values = difference(array,
                        difference_with(array, values, comparator=comparator))
    return pull_all(array, values)


def pull_at(array, *indexes):
    """Removes elements from `array` corresponding to the specified indexes and
    returns a list of the removed elements. Indexes may be specified as a list
    of indexes or as individual arguments.

    Args:
        array (list): List to pull from.
        indexes (int): Indexes to pull.

    Returns:
        list: Modified `array`.

    Warning:
        `array` is modified in place.

    Example:

        >>> pull_at([1, 2, 3, 4], 0, 2)
        [2, 4]

    .. versionadded:: 1.1.0
    """
    indexes = flatten(indexes)
    for index in sorted(indexes, reverse=True):
        del array[index]

    return array


def push(array, *items):
    """Push items onto the end of `array` and return modified `array`.

    Args:
        array (list): List to push to.
        items (mixed): Items to append.

    Returns:
        list: Modified `array`.

    Warning:
        `array` is modified in place.

    Example:

        >>> array = [1, 2, 3]
        >>> push(array, 4, 5, [6])
        [1, 2, 3, 4, 5, [6]]

    .. versionadded:: 2.2.0

    .. versionchanged:: 4.0.0
        Removed alias ``append``.
    """
    for item in items:
        array.append(item)
    return array


def remove(array, predicate=None):
    """Removes all elements from a list that the predicate returns truthy for
    and returns an array of removed elements.

    Args:
        array (list): List to remove elements from.
        predicate (mixed, optional): Predicate applied per iteration.

    Returns:
        list: Removed elements of `array`.

    Warning:
        `array` is modified in place.

    Example:

        >>> array = [1, 2, 3, 4]
        >>> items = remove(array, lambda x: x >= 3)
        >>> items
        [3, 4]
        >>> array
        [1, 2]

    .. versionadded:: 1.0.0
    """
    to_remove = [i for is_true, _, i, _ in iteriteratee(array, predicate)
                 if is_true]

    removed = []
    new_array = []

    for i, item in enumerate(array):
        if i in to_remove:
            removed.append(array[i])
        else:
            new_array.append(item)

    # Modify array in place.
    array[:] = new_array

    return removed


def reverse(array):
    """Return `array` in reverse order.

    Args:
        array (list|string): Object to process.

    Returns:
        list|string: Reverse of object.

    Example:

        >>> reverse([1, 2, 3, 4])
        [4, 3, 2, 1]

    .. versionadded:: 2.2.0
    """
    # NOTE: Using this method to reverse object since it works for both lists
    # and strings.
    return array[::-1]


def shift(array):
    """Remove the first element of `array` and return it.

    Args:
        array (list): List to shift.

    Returns:
        mixed: First element of `array`.

    Warning:
        `array` is modified in place.

    Example:

        >>> array = [1, 2, 3, 4]
        >>> item = shift(array)
        >>> item
        1
        >>> array
        [2, 3, 4]

    .. versionadded:: 2.2.0
    """
    return pop(array, 0)


def slice_(array, start=0, end=None):
    """Slices `array` from the `start` index up to, but not including, the
    `end` index.

    Args:
        array (list): Array to slice.
        start (int, optional): Start index. Defaults to ``0``.
        end (int, optional): End index. Defaults to selecting the value at
            ``start`` index.

    Returns:
        list: Sliced list.

    Example:

        >>> slice_([1, 2, 3, 4])
        [1]
        >>> slice_([1, 2, 3, 4], 1)
        [2]
        >>> slice_([1, 2, 3, 4], 1, 3)
        [2, 3]

    .. versionadded:: 1.1.0
    """
    if end is None:
        end = (start + 1) if start >= 0 else (len(array) + start + 1)

    return array[start:end]


def sort(array, comparator=None, key=None, reverse=False):
    """Sort `array` using optional `comparator`, `key`, and `reverse` options
    and return sorted `array`.

    Note:
        Python 3 removed the option to pass a custom comparator function and
        instead only allows a key function. Therefore, if a comparator
        function is passed in, it will be converted to a key function
        automatically using ``functools.cmp_to_key``.

    Args:
        array (list): List to sort.
        comparator (callable, optional): A custom comparator function used to
            sort the list. Function should accept two arguments and return a
            negative, zero, or position number depending on whether the first
            argument is considered smaller than, equal to, or larger than the
            second argument. Defaults to ``None``. This argument is mutually
            exclusive with `key`.
        key (iteratee, optional): A function of one argument used to extract a
            a comparator key from each list element. Defaults to ``None``. This
            argument is mutually exclusive with `comparator`.
        reverse (bool, optional): Whether to reverse the sort. Defaults to
            ``False``.

    Returns:
        list: Sorted list.

    Warning:
        `array` is modified in place.

    Example:

        >>> sort([2, 1, 4, 3])
        [1, 2, 3, 4]
        >>> sort([2, 1, 4, 3], reverse=True)
        [4, 3, 2, 1]
        >>> results = sort([{'a': 2, 'b': 1},\
                            {'a': 3, 'b': 2},\
                            {'a': 0, 'b': 3}],\
                           key=lambda item: item['a'])
        >>> assert results == [{'a': 0, 'b': 3},\
                               {'a': 2, 'b': 1},\
                               {'a': 3, 'b': 2}]

    .. versionadded:: 2.2.0
    """
    if comparator and key:
        raise ValueError(
            'The "comparator" and "key" arguments are mutually exclusive')

    if comparator:
        key = cmp_to_key(comparator)

    array.sort(key=key, reverse=reverse)
    return array


def sorted_index(array, value):
    """Uses a binary search to determine the lowest index at which `value`
    should be inserted into `array` in order to maintain its sort order.

    Args:
        array (list): List to inspect.
        value (mixed): Value to evaluate.

    Returns:
        int: Returns the index at which `value` should be inserted into
            `array`.

    Example:

        >>> sorted_index([1, 2, 2, 3, 4], 2)
        1

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Move iteratee support to :func:`sorted_index_by`.
    """
    return sorted_index_by(array, value)


def sorted_index_by(array, value, iteratee=None):
    """This method is like :func:`sorted_index` except that it accepts
    iteratee which is invoked for `value` and each element of `array` to
    compute their sort ranking. The iteratee is invoked with one argument:
    ``(value)``.

    Args:
        array (list): List to inspect.
        value (mixed): Value to evaluate.
        iteratee (mixed, optional): The iteratee invoked per element. Defaults
            to :func:`.identity`.

    Returns:
        int: Returns the index at which `value` should be inserted into
            `array`.

    Example:

        >>> array = [{'x': 4}, {'x': 5}]
        >>> sorted_index_by(array, {'x': 4}, lambda o: o['x'])
        0
        >>> sorted_index_by(array, {'x': 4}, 'x')
        0

    .. versionadded:: 4.0.0
    """
    if iteratee:
        # Generate array of sorted keys computed using iteratee.
        iteratee = pyd.iteratee(iteratee)
        array = sorted(iteratee(item) for item in array)
        value = iteratee(value)

    return bisect_left(array, value)


def sorted_index_of(array, value):
    """Returns the index of the matched `value` from the sorted `array`, else
    ``-1``.

    Args:
        array (list): Array to inspect.
        value (mixed): Value to search for.

    Returns:
        int: Returns the index of the first matched value, else ``-1``.

    Example:

        >>> sorted_index_of([3, 5, 7, 10], 3)
        0
        >>> sorted_index_of([10, 10, 5, 7, 3], 10)
        -1

    .. versionadded:: 4.0.0
    """
    index = sorted_index(array, value)

    if index < len(array) and array[index] == value:
        return index
    else:
        return -1


def sorted_last_index(array, value):
    """This method is like :func:`sorted_index` except that it returns the
    highest index at which `value` should be inserted into `array` in order to
    maintain its sort order.

    Args:
        array (list): List to inspect.
        value (mixed): Value to evaluate.

    Returns:
        int: Returns the index at which `value` should be inserted into
            `array`.

    Example:

        >>> sorted_last_index([1, 2, 2, 3, 4], 2)
        3

    .. versionadded:: 1.1.0

    .. versionchanged:: 4.0.0
        Move iteratee support to :func:`sorted_last_index_by`.
    """
    return sorted_last_index_by(array, value)


def sorted_last_index_by(array, value, iteratee=None):
    """This method is like :func:`sorted_last_index` except that it accepts
    iteratee which is invoked for `value` and each element of `array` to
    compute their sort ranking. The iteratee is invoked with one argument:
    ``(value)``.

    Args:
        array (list): List to inspect.
        value (mixed): Value to evaluate.
        iteratee (mixed, optional): The iteratee invoked per element. Defaults
            to :func:`.identity`.

    Returns:
        int: Returns the index at which `value` should be inserted into
            `array`.

    Example:

        >>> array = [{'x': 4}, {'x': 5}]
        >>> sorted_last_index_by(array, {'x': 4}, lambda o: o['x'])
        1
        >>> sorted_last_index_by(array, {'x': 4}, 'x')
        1
    """
    if iteratee:
        # Generate array of sorted keys computed using iteratee.
        iteratee = pyd.iteratee(iteratee)
        array = sorted(iteratee(item) for item in array)
        value = iteratee(value)

    return bisect_right(array, value)


def sorted_last_index_of(array, value):
    """This method is like :func:`last_index_of` except that it performs a
    binary search on a sorted `array`.

    Args:
        array (list): Array to inspect.
        value (mixed): Value to search for.

    Returns:
        int: Returns the index of the matched value, else ``-1``.

    Example:

        >>> sorted_last_index_of([4, 5, 5, 5, 6], 5)
        3
        >>> sorted_last_index_of([6, 5, 5, 5, 4], 6)
        -1

    .. versionadded:: 4.0.0
    """
    index = sorted_last_index(array, value) - 1

    if index < len(array) and array[index] == value:
        return index
    else:
        return -1


def sorted_uniq(array):
    """Return sorted array with unique elements.

    Args:
        array (list): List of values to be sorted.

    Returns:
        list: List of unique elements in a sorted fashion.

    Example:

        >>> sorted_uniq([4, 2, 2, 5])
        [2, 4, 5]
        >>> sorted_uniq([-2, -2, 4, 1])
        [-2, 1, 4]

    .. versionadded:: 4.0.0
    """
    return sorted(uniq(array))


def sorted_uniq_by(array, iteratee=None):
    """This method is like :func:`sorted_uniq` except that it accepts iteratee
    which is invoked for each element in array to generate the criterion by
    which uniqueness is computed. The order of result values is determined by
    the order they occur in the array. The iteratee is invoked with one
    argument: ``(value)``.

    Args:
        array (list): List of values to be sorted.
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: Unique list.

    Example:

        >>> sorted_uniq_by([3, 2, 1, 3, 2, 1], lambda val: val % 2)
        [2, 3]

    .. versionadded:: 4.0.0
    """
    return sorted(uniq_by(array, iteratee=iteratee))


def splice(array, start, count=None, *items):
    """Modify the contents of `array` by inserting elements starting at index
    `start` and removing `count` number of elements after.

    Args:
        array (list|str): List to splice.
        start (int): Start to splice at.
        count (int, optional): Number of items to remove starting at
            `start`. If ``None`` then all items after `start` are removed.
            Defaults to ``None``.
        items (mixed): Elements to insert starting at `start`. Each item is
            inserted in the order given.

    Returns:
        list|str: The removed elements of `array` or the spliced string.

    Warning:
        `array` is modified in place if ``list``.

    Example:

        >>> array = [1, 2, 3, 4]
        >>> splice(array, 1)
        [2, 3, 4]
        >>> array
        [1]
        >>> array = [1, 2, 3, 4]
        >>> splice(array, 1, 2)
        [2, 3]
        >>> array
        [1, 4]
        >>> array = [1, 2, 3, 4]
        >>> splice(array, 1, 2, 0, 0)
        [2, 3]
        >>> array
        [1, 0, 0, 4]

    .. versionadded:: 2.2.0

    .. versionchanged:: 3.0.0
        Support string splicing.
    """
    if count is None:
        count = len(array) - start

    is_string = pyd.is_string(array)

    if is_string:
        array = list(array)

    removed = array[start:start + count]
    del array[start:start + count]

    for item in reverse(items):
        array.insert(start, item)

    if is_string:
        return ''.join(array)
    else:
        return removed


def split_at(array, index):
    """Returns a list of two lists composed of the split of `array` at `index`.

    Args:
        array (list): List to split.
        index (int): Index to split at.

    Returns:
        list: Split list.

    Example:

        >>> split_at([1, 2, 3, 4], 2)
        [[1, 2], [3, 4]]

    .. versionadded:: 2.0.0
    """
    return [take(array, index), drop(array, index)]


def tail(array):
    """Return all but the first element of `array`.

    Args:
        array (list): List to process.

    Returns:
        list: Rest of the list.

    Example:

        >>> tail([1, 2, 3, 4])
        [2, 3, 4]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Renamed from ``rest`` to ``tail``.
    """
    return array[1:]


def take(array, n=1):
    """Creates a slice of `array` with `n` elements taken from the beginning.

    Args:
        array (list): List to process.
        n (int, optional): Number of elements to take. Defaults to ``1``.

    Returns:
        list: Taken list.

    Example:

        >>> take([1, 2, 3, 4], 2)
        [1, 2]

    .. versionadded:: 1.0.0

    .. versionchanged:: 1.1.0
        Added ``n`` argument and removed as alias of :func:`first`.

    .. versionchanged:: 3.0.0
        Made ``n`` default to ``1``.
    """
    return take_while(array, lambda _, index: index < n)


def take_right(array, n=1):
    """Creates a slice of `array` with `n` elements taken from the end.

    Args:
        array (list): List to process.
        n (int, optional): Number of elements to take. Defaults to ``1``.

    Returns:
        list: Taken list.

    Example:

        >>> take_right([1, 2, 3, 4], 2)
        [3, 4]

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Made ``n`` default to ``1``.
    """
    length = len(array)
    return take_right_while(array, lambda _, index: (length - index) <= n)


def take_right_while(array, predicate=None):
    """Creates a slice of `array` with elements taken from the end. Elements
    are taken until the `predicate` returns falsey. The `predicate` is
    invoked with three arguments: ``(value, index, array)``.

    Args:
        array (list): List to process.
        predicate (mixed): Predicate called per iteration

    Returns:
        list: Dropped list.

    Example:

        >>> take_right_while([1, 2, 3, 4], lambda x: x >= 3)
        [3, 4]

    .. versionadded:: 1.1.0
    """
    n = len(array)
    for is_true, _, _, _ in iteriteratee(array, predicate, reverse=True):
        if is_true:
            n -= 1
        else:
            break

    return array[n:]


def take_while(array, predicate=None):
    """Creates a slice of `array` with elements taken from the beginning.
    Elements are taken until the `predicate` returns falsey. The
    `predicate` is invoked with three arguments: ``(value, index, array)``.

    Args:
        array (list): List to process.
        predicate (mixed): Predicate called per iteration

    Returns:
        list: Taken list.

    Example:

        >>> take_while([1, 2, 3, 4], lambda x: x < 3)
        [1, 2]

    .. versionadded:: 1.1.0
    """
    n = 0
    for is_true, _, _, _ in iteriteratee(array, predicate):
        if is_true:
            n += 1
        else:
            break

    return array[:n]


def union(array, *others):
    """Computes the union of the passed-in arrays.

    Args:
        array (list): List to union with.
        others (list): Lists to unionize with `array`.

    Returns:
        list: Unionized list.

    Example:

        >>> union([1, 2, 3], [2, 3, 4], [3, 4, 5])
        [1, 2, 3, 4, 5]

    .. versionadded:: 1.0.0
    """
    if not others:
        return array[:]

    return uniq(flatten([array] + list(others)))


def union_by(array, *others, **kargs):
    """This method is similar to :func:`union` except that it accepts iteratee
    which is invoked for each element of each arrays to generate the criterion
    by which uniqueness is computed.

    Args:
        array (list): List to unionize with.
        others (list): Lists to unionize with `array`.

    Keyword Args:
        iteratee (function): Function to invoke on each element.

    Returns:
        list: Unionized list.

    Example:

        >>> union_by([1, 2, 3], [2, 3, 4], iteratee=lambda x: x % 2)
        [1, 2]
        >>> union_by([1, 2, 3], [2, 3, 4], iteratee=lambda x: x % 9)
        [1, 2, 3, 4]

    .. versionadded:: 4.0.0
    """
    if not others:
        return array[:]

    iteratee, others = parse_iteratee('iteratee', *others, **kargs)

    return uniq_by(flatten([array] + list(others)), iteratee=iteratee)


def union_with(array, *others, **kargs):
    """This method is like :func:`union` except that it accepts comparator
    which is invoked to compare elements of arrays. Result values are chosen
    from the first array in which the value occurs.

    Args:
        array (list): List to unionize with.
        others (list): Lists to unionize with `array`.

    Keyword Args:
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: Unionized list.

    Example:

        >>> comparator = lambda a, b: (a % 2) == (b % 2)
        >>> union_with([1, 2, 3], [2, 3, 4], comparator=comparator)
        [1, 2]
        >>> union_with([1, 2, 3], [2, 3, 4])
        [1, 2, 3, 4]

    .. versionadded:: 4.0.0
    """
    if not others:
        return array[:]

    comparator, others = parse_iteratee('comparator', *others, **kargs)

    return uniq_with(flatten([array] + list(others)), comparator=comparator)


def uniq(array):
    """Creates a duplicate-value-free version of the array. If iteratee is
    passed, each element of array is passed through a iteratee before
    uniqueness is computed. The iteratee is invoked with three arguments:
    ``(value, index, array)``. If an object path is passed for iteratee, the
    created iteratee will return the path value of the given element. If an
    object is passed for iteratee, the created filter style iteratee will
    return ``True`` for elements that have the properties of the given object,
    else ``False``.

    Args:
        array (list): List to process.

    Returns:
        list: Unique list.

    Example:

        >>> uniq([1, 2, 3, 1, 2, 3])
        [1, 2, 3]

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0

        - Moved `iteratee` argument to :func:`uniq_by`.
        - Removed alias ``unique``.

    """
    return uniq_by(array)


def uniq_by(array, iteratee=None):
    """This method is like :func:`uniq` except that it accepts iteratee which
    is invoked for each element in array to generate the criterion by which
    uniqueness is computed. The order of result values is determined by the
    order they occur in the array. The iteratee is invoked with one argument:
    ``(value)``.

    Args:
        array (list): List to process.
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: Unique list.

    Example:

        >>> uniq_by([1, 2, 3, 1, 2, 3], lambda val: val % 2)
        [1, 2]

    .. versionadded:: 4.0.0
    """
    return list(iterunique(array, iteratee=iteratee))


def uniq_with(array, comparator=None):
    """This method is like _.uniq except that it accepts comparator which is
    invoked to compare elements of array. The order of result values is
    determined by the order they occur in the array.The comparator is invoked
    with two arguments: ``(value, other)``.

    Args:
        array (list): List to process.
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: Unique list.

    Example:

        >>> uniq_with([1, 2, 3, 4, 5], lambda a, b: (a % 2) == (b % 2))
        [1, 2]

    .. versionadded:: 4.0.0
    """
    return list(iterunique(array, comparator=comparator))


def unshift(array, *items):
    """Insert the given elements at the beginning of `array` and return the
    modified list.

    Args:
        array (list): List to modify.
        items (mixed): Items to insert.

    Returns:
        list: Modified list.

    Warning:
        `array` is modified in place.

    Example:

        >>> array = [1, 2, 3, 4]
        >>> unshift(array, -1, -2)
        [-1, -2, 1, 2, 3, 4]
        >>> array
        [-1, -2, 1, 2, 3, 4]

    .. versionadded:: 2.2.0
    """
    for item in reverse(items):
        array.insert(0, item)

    return array


def unzip(array):
    """The inverse of :func:`zip_`, this method splits groups of
    elements into lists composed of elements from each group at their
    corresponding indexes.

    Args:
        array (list): List to process.

    Returns:
        list: Unzipped list.

    Example:

        >>> unzip([[1, 4, 7], [2, 5, 8], [3, 6, 9]])
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    .. versionadded:: 1.0.0
    """
    return zip_(*array)


def unzip_with(array, iteratee=None):
    """This method is like :func:`unzip` except that it accepts a iteratee to
    specify how regrouped values should be combined. The iteratee is invoked
    with four arguments: ``(accumulator, value, index, group)``.

    Args:
        array (list): List to process.
        iteratee (callable, optional): Function to combine regrouped values.

    Returns:
        list: Unzipped list.

    Example:

        >>> from pydash import add
        >>> unzip_with([[1, 10, 100], [2, 20, 200]], add)
        [3, 30, 300]

    .. versionadded:: 3.3.0
    """
    if not array:
        return []

    result = unzip(array)

    if iteratee is None:
        return result

    def cbk(group):
        return pyd.reduce_(group, iteratee)

    return pyd.map_(result, cbk)


def without(array, *values):
    """Creates an array with all occurrences of the passed values removed.

    Args:
        array (list): List to filter.
        values (mixed): Values to remove.

    Returns:
        list: Filtered list.

    Example:

        >>> without([1, 2, 3, 2, 4, 4], 2, 4)
        [1, 3]

    .. versionadded:: 1.0.0
    """
    return [item for item in array if item not in values]


def xor(array, *lists):
    """Creates a list that is the symmetric difference of the provided lists.

    Args:
        array (list): List to process.
        *lists (list): Lists to xor with.

    Returns:
        list: XOR'd list.

    Example:

        >>> xor([1, 3, 4], [1, 2, 4], [2])
        [3]

    .. versionadded:: 1.0.0
    """
    return xor_by(array, *lists)


def xor_by(array, *lists, **kargs):
    """This method is like :func:`xor` except that it accepts iteratee which is
    invoked for each element of each arrays to generate the criterion by which
    by which they're compared. The order of result values is determined by the
    order they occur in the arrays. The iteratee is invoked with one argument:
    ``(value)``.

    Args:
        array (list): List to process.
        *lists (list): Lists to xor with.

    Keyword Args:
        iteratee (mixed, optional): Function to transform the elements of the
            arrays. Defaults to :func:`.identity`.

    Returns:
        list: XOR'd list.

    Example:

        >>> xor_by([2.1, 1.2], [2.3, 3.4], round)
        [1.2, 3.4]
        >>> xor_by([{'x': 1}], [{'x': 2}, {'x': 1}], 'x')
        [{'x': 2}]

    .. versionadded:: 4.0.0
    """
    if not lists:
        return array[:]

    iteratee, lists = parse_iteratee('iteratee', *lists, **kargs)

    return xor(uniq(difference_by(array + lists[0],
                                  intersection_by(array, lists[0],
                                                  iteratee=iteratee),
                                  iteratee=iteratee)),
               *lists[1:])


def xor_with(array, *lists, **kargs):
    """This method is like :func:`xor` except that it accepts comparator which
    is invoked to compare elements of arrays. The order of result values is
    determined by the order they occur in the arrays. The comparator is invoked
    with two arguments: ``(arr_val, oth_val)``.

    Args:
        array (list): List to process.
        *lists (list): Lists to xor with.

    Keyword Args:
        comparator (callable, optional): Function to compare the elements of
            the arrays. Defaults to :func:`.is_equal`.

    Returns:
        list: XOR'd list.

    Example:

        >>> objects = [{'x': 1, 'y': 2}, {'x': 2, 'y': 1}]
        >>> others = [{'x': 1, 'y': 1}, {'x': 1, 'y': 2}]
        >>> expected = [{'y': 1, 'x': 2}, {'y': 1, 'x': 1}]
        >>> xor_with(objects, others, lambda a, b: a == b) == expected
        True

    .. versionadded:: 4.0.0
    """
    if not lists:
        return array[:]

    comp, lists = parse_iteratee('comparator', *lists, **kargs)

    return xor_with(uniq(difference_with(array + lists[0],
                                         intersection_with(array, lists[0],
                                                           comparator=comp),
                                         comparator=comp)),
                    *lists[1:])


def zip_(*arrays):
    """Groups the elements of each array at their corresponding indexes.
    Useful for separate data sources that are coordinated through matching
    array indexes.

    Args:
        arrays (list): Lists to process.

    Returns:
        list: Zipped list.

    Example:

        >>> zip_([1, 2, 3], [4, 5, 6], [7, 8, 9])
        [[1, 4, 7], [2, 5, 8], [3, 6, 9]]

    .. versionadded:: 1.0.0
    """
    # zip returns as a list of tuples so convert to list of lists
    return [list(item) for item in zip(*arrays)]


def zip_object(keys, values=None):
    """Creates a dict composed from lists of keys and values. Pass either a
    single two dimensional list, i.e. ``[[key1, value1], [key2, value2]]``, or
    two lists, one of keys and one of corresponding values.

    Args:
        keys (list): Either a list of keys or a list of ``[key, value]`` pairs.
        values (list, optional): List of values to zip.

    Returns:
        dict: Zipped dict.

    Example:

        >>> zip_object([1, 2, 3], [4, 5, 6])
        {1: 4, 2: 5, 3: 6}

    .. versionadded:: 1.0.0

    .. versionchanged:: 4.0.0
        Removed alias ``object_``.
    """

    if values is None:
        keys, values = unzip(keys)

    return dict(zip(keys, values))


def zip_object_deep(keys, values=None):
    """This method is like :func:`zip_object` except that it supports property
    paths.

    Args:
        keys (list): Either a list of keys or a list of ``[key, value]`` pairs.
        values (list, optional): List of values to zip.

    Returns:
        dict: Zipped dict.

    Example:

        >>> expected = {'a': {'b': {'c': 1, 'd': 2}}}
        >>> zip_object_deep(['a.b.c', 'a.b.d'], [1, 2]) == expected
        True

    .. versionadded:: 4.0.0
    """
    if values is None:  # pragma: no cover
        keys, values = unzip(keys)

    obj = {}
    for idx, key in enumerate(keys):
        obj = pyd.set_(obj, key, pyd.get(values, idx))

    return obj


def zip_with(*arrays, **kargs):
    """This method is like :func:`zip` except that it accepts a iteratee to
    specify how grouped values should be combined. The iteratee is invoked with
    four arguments: ``(accumulator, value, index, group)``.

    Args:
        *arrays (list): Lists to process.

    Keyword Args:
        iteratee (function): Function to combine grouped values.

    Returns:
        list: Zipped list of grouped elements.

    Example:

        >>> from pydash import add
        >>> zip_with([1, 2], [10, 20], [100, 200], add)
        [111, 222]
        >>> zip_with([1, 2], [10, 20], [100, 200], iteratee=add)
        [111, 222]

    .. versionadded:: 3.3.0
    """
    if 'iteratee' in kargs:
        iteratee = kargs['iteratee']
    elif(len(arrays) > 1):
        iteratee = arrays[-1]
        arrays = arrays[:-1]
    else:
        iteratee = None

    return unzip_with(arrays, iteratee)


#
# Utility methods not a part of the main API
#


def iterflatten(array, depth=-1):
    """Iteratively flatten a list shallowly or deeply."""
    for item in array:
        if isinstance(item, (list, tuple)) and depth != 0:
            for subitem in iterflatten(item, depth - 1):
                yield subitem
        else:
            yield item


def iterinterleave(*arrays):
    """Interleave multiple lists."""
    iters = [iter(arr) for arr in arrays]

    while iters:
        nextiters = []
        for itr in iters:
            try:
                yield next(itr)
                nextiters.append(itr)
            except StopIteration:
                pass

        iters = nextiters


def iterintersperse(iterable, separator):
    """Iteratively intersperse iterable."""
    iterable = iter(iterable)
    yield next(iterable)
    for item in iterable:
        yield separator
        yield item


def iterunique(array, comparator=None, iteratee=None):
    """Yield each unique item in array."""
    if not array:  # pragma: no cover
        return

    if comparator is None:
        comparator = pyd.is_equal

    iteratee = pyd.iteratee(iteratee)

    seen = []
    for item in array:
        cmp_item = iteratee(item)
        new = True

        for seen_item in seen:
            if comparator(cmp_item, seen_item):
                new = False
                break

        if new:
            yield item
            seen.append(cmp_item)


def iterduplicates(array):
    """Yield duplictes found in `array`."""
    seen = []
    for i, item in enumerate(array):
        if item in seen:
            yield (i, item)
        else:
            seen.append(item)


def iterintersection(array, other, comparator=None, iteratee=None):
    """Yield intersecting values between `array` and `other` using `comparator`
    to determine if they intersect.
    """
    if not array or not other:  # pragma: no cover
        return

    if comparator is None:
        comparator = pyd.is_equal

    iteratee = pyd.iteratee(iteratee)

    # NOTE: Maintain ordering of yielded values based on `array` ordering.
    seen = []
    for item in array:
        cmp_item = iteratee(item)

        if cmp_item in seen:
            continue

        seen.append(cmp_item)
        seen_others = []

        for value in other:
            cmp_value = iteratee(value)

            if cmp_value in seen_others:
                continue

            seen_others.append(cmp_value)

            if comparator(cmp_item, cmp_value):
                yield item
                break


def iterdifference(array, other, comparator=None, iteratee=None):
    """Yield different values in `array` as compared to `other` using
    `comparator` to determine if they are different.
    """
    if not array or not other:  # pragma: no cover
        return

    if comparator is None:
        comparator = pyd.is_equal

    iteratee = pyd.iteratee(iteratee)

    def is_different(item, seen):
        is_diff = True

        if item not in seen:
            for value in other:
                if comparator(iteratee(item), iteratee(value)):
                    is_diff = False
                    break

            if is_diff:
                seen.append(item)
        return is_diff

    seen = []
    not_seen = []

    for item in array:
        if item in not_seen or is_different(item, seen):
            yield item



# -*- coding: utf-8 -*-
"""Functions that wrap other functions.

.. versionadded:: 1.0.0
"""

from __future__ import absolute_import

import time

import pydash as pyd
from ._compat import _range, getfullargspec


__all__ = (
    'after',
    'ary',
    'before',
    'conjoin',
    'curry',
    'curry_right',
    'debounce',
    'delay',
    'disjoin',
    'flip',
    'flow',
    'flow_right',
    'iterated',
    'juxtapose',
    'negate',
    'once',
    'over_args',
    'partial',
    'partial_right',
    'rearg',
    'spread',
    'throttle',
    'unary',
    'wrap',
)


class After(object):
    """Wrap a function in an after context."""
    def __init__(self, func, n):
        try:
            n = int(n)
            assert n >= 0
        except (ValueError, TypeError, AssertionError):
            n = 0

        self.n = n
        self.func = func

    def __call__(self, *args, **kargs):
        """Return results of :attr:`func` after :attr:`n` calls."""
        self.n -= 1

        if self.n <= 0:
            return self.func(*args, **kargs)


class Ary(object):
    """Wrap a function in an ary context."""
    def __init__(self, func, n):
        try:
            n = int(n)
            assert n >= 0
        except (ValueError, TypeError, AssertionError):
            n = None

        self.n = n
        self.func = func

    def __call__(self, *args, **kargs):
        """Return results of :attr:`func` with arguments capped to :attr:`n`.
        Only positional arguments are capped. Any number of keyword arguments
        are allowed.
        """
        if self.n is not None:
            args = args[:self.n]

        return self.func(*args, **kargs)


class Before(After):
    """Wrap a function in a before context."""
    def __call__(self, *args, **kargs):
        self.n -= 1

        if self.n > 0:
            return self.func(*args, **kargs)


class Flow(object):
    """Wrap a function in a flow context."""
    def __init__(self, *funcs, **kargs):
        self.funcs = funcs
        self.from_right = kargs.get('from_right', True)

    def __call__(self, *args, **kargs):
        """Return results of composing :attr:`funcs`."""
        funcs = list(self.funcs)
        from_index = -1 if self.from_right else 0

        result = None

        while funcs:
            result = funcs.pop(from_index)(*args, **kargs)
            args = (result,)
            kargs = {}

        return result


class Conjoin(object):
    """Wrap a set of functions in a conjoin context."""
    def __init__(self, *funcs):
        self.funcs = funcs

    def __call__(self, obj):
        """Return result of conjoin `obj` with :attr:`funcs` predicates."""
        def iteratee(item):
            return pyd.every(self.funcs, lambda func: func(item))

        return pyd.every(obj, iteratee)


class Curry(object):
    """Wrap a function in a curry context."""
    def __init__(self, func, arity, args=None, kargs=None):
        self.func = func
        self.arity = (len(getfullargspec(func).args) if arity is None
                      else arity)
        self.args = () if args is None else args
        self.kargs = {} if kargs is None else kargs

    def __call__(self, *args, **kargs):
        """Store `args` and `kargs` and call :attr:`func` if we've reached or
        exceeded the function arity.
        """
        args = self.compose_args(args)
        kargs.update(self.kargs)

        if (len(args) + len(kargs)) >= self.arity:
            args_arity = self.arity - len(kargs)
            args = args[:(args_arity if args_arity > 0 else 0)]
            curried = self.func(*args, **kargs)
        else:
            # NOTE: Use self.__class__ so that subclasses will use their own
            # class to generate next iteration of call.
            curried = self.__class__(self.func, self.arity, args, kargs)

        return curried

    def compose_args(self, new_args):
        """Combine `self.args` with `new_args` and return."""
        return tuple(list(self.args) + list(new_args))


class CurryRight(Curry):
    """Wrap a function in a curry-right context."""
    def compose_args(self, new_args):
        return tuple(list(new_args) + list(self.args))


class Debounce(object):
    """Wrap a function in a debounce context."""
    def __init__(self, func, wait, max_wait=False):
        self.func = func
        self.wait = wait
        self.max_wait = max_wait

        self.last_result = None

        # Initialize last_* times to be prior to the wait periods so that func
        # is primed to be executed on first call.
        self.last_call = pyd.now() - self.wait
        self.last_execution = (pyd.now() - max_wait if pyd.is_number(max_wait)
                               else None)

    def __call__(self, *args, **kargs):
        """Execute :attr:`func` if function hasn't been called witinin last
        :attr:`wait` milliseconds or in last :attr:`max_wait` milliseconds.
        Return results of last successful call.
        """
        present = pyd.now()

        if any([(present - self.last_call) >= self.wait,
                (self.max_wait and
                 (present - self.last_execution) >= self.max_wait)]):
            self.last_result = self.func(*args, **kargs)
            self.last_execution = present

        self.last_call = present

        return self.last_result


class Disjoin(object):
    """Wrap a set of functions in a disjoin context."""
    def __init__(self, *funcs):
        self.funcs = funcs

    def __call__(self, obj):
        """Return result of disjoin `obj` with :attr:`funcs` predicates."""
        def iteratee(item):
            return pyd.some(self.funcs, lambda func: func(item))

        return pyd.some(obj, iteratee)


class Flip(object):
    """Wrap a function in a flip context."""
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        return self.func(*reversed(args), **kargs)


class Iterated(object):
    """Wrap a function in an iterated context."""
    def __init__(self, func):
        self.func = func

    def _iteration(self, initial):
        """Iterator that composing :attr:`func` with itself."""
        value = initial
        while True:
            value = self.func(value)
            yield value

    def __call__(self, initial, n):
        """Return value of calling :attr:`func` `n` times using `initial` as
        seed value.
        """
        value = initial
        iteration = self._iteration(value)

        for _ in _range(n):
            value = next(iteration)

        return value


class Juxtapose(object):
    """Wrap a function in a juxtapose context."""
    def __init__(self, *funcs):
        self.funcs = funcs

    def __call__(self, *objs):
        return pyd.map_(self.funcs, lambda func: func(*objs))


class OverArgs(object):
    """Wrap a function in a over_args context."""
    def __init__(self, func, *transforms):
        self.func = func
        self.transforms = pyd.flatten(transforms)

    def __call__(self, *args):
        args = (self.transforms[idx](args) for idx, args in enumerate(args))
        return self.func(*args)


class Negate(object):
    """Wrap a function in a negate context."""
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        """Return negated results of calling :attr:`func`."""
        return not self.func(*args, **kargs)


class Once(object):
    """Wrap a function in a once context."""
    def __init__(self, func):
        self.func = func
        self.result = None
        self.called = False

    def __call__(self, *args, **kargs):
        """Return results from the first call of :attr:`func`."""
        if not self.called:
            self.result = self.func(*args, **kargs)
            self.called = True

        return self.result


class Partial(object):
    """Wrap a function in a partial context."""
    def __init__(self, func, args, kargs=None, from_right=False):
        self.func = func
        self.args = args
        self.kargs = kargs or {}
        self.from_right = from_right

    def __call__(self, *args, **kargs):
        """Return results from :attr:`func` with :attr:`args` + `args`. Apply
        arguments from left or right depending on :attr:`from_right`.
        """
        if self.from_right:
            args = list(args) + list(self.args)
        else:
            args = list(self.args) + list(args)

        kargs = dict(list(self.kargs.items()) + list(kargs.items()))

        return self.func(*args, **kargs)


class Rearg(object):
    """Wrap a function in a rearg context."""
    def __init__(self, func, *indexes):
        self.func = func

        # Index `indexes` by the index value so we can do a lookup mapping by
        # walking the function arguments.
        self.indexes = dict(
            (src_index, dest_index)
            for dest_index, src_index in enumerate(pyd.flatten(indexes)))

    def __call__(self, *args, **kargs):
        """Return results from :attr:`func` using rearranged arguments."""
        reargs = {}
        rest = []

        # Walk arguments to ensure each one is added to the final argument
        # list.
        for src_index, arg in enumerate(args):
            # NOTE: dest_index will range from 0 to len(indexes).
            dest_index = self.indexes.get(src_index)

            if dest_index is not None:
                # Remap argument index.
                reargs[dest_index] = arg
            else:
                # Argumnet index is not contained in `indexes` so stick in the
                # back.
                rest.append(arg)

        reargs = [reargs[key] for key in sorted(reargs)] + rest

        return self.func(*reargs, **kargs)


class Spread(object):
    """Wrap a function in a spread context."""
    def __init__(self, func):
        self. func = func

    def __call__(self, args):
        """Return results from :attr:`func` using array of `args` provided."""
        return self.func(args)


class Throttle(object):
    """Wrap a function in a throttle context."""
    def __init__(self, func, wait):
        self.func = func
        self.wait = wait

        self.last_result = None
        self.last_execution = pyd.now() - self.wait

    def __call__(self, *args, **kargs):
        """Execute :attr:`func` if function hasn't been called witinin last
        :attr:`wait` milliseconds. Return results of last successful call.
        """
        present = pyd.now()

        if (present - self.last_execution) >= self.wait:
            self.last_result = self.func(*args, **kargs)
            self.last_execution = present

        return self.last_result


def after(func, n):
    """Creates a function that executes `func`, with the arguments of the
    created function, only after being called `n` times.

    Args:
        func (function): Function to execute.
        n (int): Number of times `func` must be called before it is executed.

    Returns:
        After: Function wrapped in an :class:`After` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> after_func = after(func, 3)
        >>> after_func(1, 2, 3)
        >>> after_func(1, 2, 3)
        >>> after_func(1, 2, 3)
        (1, 2, 3)
        >>> after_func(4, 5, 6)
        (4, 5, 6)

    .. versionadded:: 1.0.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `func` first.
    """
    return After(func, n)


def ary(func, n):
    """Creates a function that accepts up to `n` arguments ignoring any
    additional arguments. Only positional arguments are capped. All keyword
    arguments are allowed through.

    Args:
        func (function): Function to cap arguments for.
        n (int): Number of arguments to accept.

    Returns:
        Ary: Function wrapped in an :class:`Ary` context.

    Example:

        >>> func = lambda a, b, c=0, d=5: (a, b, c, d)
        >>> ary_func = ary(func, 2)
        >>> ary_func(1, 2, 3, 4, 5, 6)
        (1, 2, 0, 5)
        >>> ary_func(1, 2, 3, 4, 5, 6, c=10, d=20)
        (1, 2, 10, 20)

    .. versionadded:: 3.0.0
    """
    return Ary(func, n)


def before(func, n):
    """Creates a function that executes `func`, with the arguments of the
    created function, until it has been called `n` times.

    Args:
        func (function): Function to execute.
        n (int): Number of times `func` may be executed.

    Returns:
        Before: Function wrapped in an :class:`Before` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> before_func = before(func, 3)
        >>> before_func(1, 2, 3)
        (1, 2, 3)
        >>> before_func(1, 2, 3)
        (1, 2, 3)
        >>> before_func(1, 2, 3)
        >>> before_func(1, 2, 3)

    .. versionadded:: 1.1.0

    .. versionchanged:: 3.0.0
        Reordered arguments to make `func` first.
    """
    return Before(func, n)


def conjoin(*funcs):
    """Creates a function that composes multiple predicate functions into a
    single predicate that tests whether **all** elements of an object pass each
    predicate.

    Args:
        *funcs (function): Function(s) to conjoin.

    Returns:
        Conjoin: Function(s) wrapped in a :class:`Conjoin` context.

    Example:

        >>> conjoiner = conjoin(lambda x: isinstance(x, int), lambda x: x > 3)
        >>> conjoiner([1, 2, 3])
        False
        >>> conjoiner([1.0, 2, 1])
        False
        >>> conjoiner([4.0, 5, 6])
        False
        >>> conjoiner([4, 5, 6])
        True

    .. versionadded:: 2.0.0
    """
    return Conjoin(*funcs)


def curry(func, arity=None):
    """Creates a function that accepts one or more arguments of `func` that
    when invoked either executes `func` returning its result (if all `func`
    arguments have been provided) or returns a function that accepts one or
    more of the remaining `func` arguments, and so on.

    Args:
        func (function): Function to curry.
        arity (int, optional): Number of function arguments that can be
            accepted by curried function. Default is to use the number of
            arguments that are accepted by `func`.

    Returns:
        Curry: Function wrapped in a :class:`Curry` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> currier = curry(func)
        >>> currier = currier(1)
        >>> assert isinstance(currier, Curry)
        >>> currier = currier(2)
        >>> assert isinstance(currier, Curry)
        >>> currier = currier(3)
        >>> currier
        (1, 2, 3)

    .. versionadded:: 1.0.0
    """
    return Curry(func, arity)


def curry_right(func, arity=None):
    """This method is like :func:`curry` except that arguments are applied to
    `func` in the manner of :func:`partial_right` instead of :func:`partial`.

    Args:
        func (function): Function to curry.
        arity (int, optional): Number of function arguments that can be
            accepted by curried function. Default is to use the number of
            arguments that are accepted by `func`.

    Returns:
        CurryRight: Function wrapped in a :class:`CurryRight` context.

    Example:

        >>> func = lambda a, b, c: (a, b, c)
        >>> currier = curry_right(func)
        >>> currier = currier(1)
        >>> assert isinstance(currier, CurryRight)
        >>> currier = currier(2)
        >>> assert isinstance(currier, CurryRight)
        >>> currier = currier(3)
        >>> currier
        (3, 2, 1)

    .. versionadded:: 1.1.0
    """
    return CurryRight(func, arity)


def debounce(func, wait, max_wait=False):
    """Creates a function that will delay the execution of `func` until after
    `wait` milliseconds have elapsed since the last time it was invoked.
    Subsequent calls to the debounced function will return the result of the
    last `func` call.

    Args:
        func (function): Function to execute.
        wait (int): Milliseconds to wait before executing `func`.
        max_wait (optional): Maximum time to wait before executing `func`.

    Returns:
        Debounce: Function wrapped in a :class:`Debounce` context.

    .. versionadded:: 1.0.0
    """
    return Debounce(func, wait, max_wait=max_wait)


def delay(func, wait, *args, **kargs):
    """Executes the `func` function after `wait` milliseconds. Additional
    arguments will be provided to `func` when it is invoked.

    Args:
        func (function): Function to execute.
        wait (int): Milliseconds to wait before executing `func`.
        *args (optional): Arguments to pass to `func`.
        **kargs (optional): Keyword arguments to pass to `func`.

    Returns:
        mixed: Return from `func`.

    .. versionadded:: 1.0.0
    """
    time.sleep(wait / 1000.0)
    return func(*args, **kargs)


def disjoin(*funcs):
    """Creates a function that composes multiple predicate functions into a
    single predicate that tests whether **any** elements of an object pass each
    predicate.

    Args:
        *funcs (function): Function(s) to disjoin.

    Returns:
        Disjoin: Function(s) wrapped in a :class:`Disjoin` context.

    Example:

        >>> disjoiner = disjoin(lambda x: isinstance(x, float),\
                                lambda x: isinstance(x, int))
        >>> disjoiner([1, '2', '3'])
        True
        >>> disjoiner([1.0, '2', '3'])
        True
        >>> disjoiner(['1', '2', '3'])
        False

    .. versionadded:: 2.0.0
    """
    return Disjoin(*funcs)


def flip(func):
    """Creates a function that invokes the method with arguments reversed.

    Args:
        func (function): Function to flip arguments for.

    Returns:
        function: Function wrapped in a :class:`Flip` context.

    Example:

        >>> flipped = flip(lambda *args: args)
        >>> flipped(1, 2, 3, 4)
        (4, 3, 2, 1)
        >>> flipped = flip(lambda *args: [i * 2 for i in args])
        >>> flipped(1, 2, 3, 4)
        [8, 6, 4, 2]

    .. versionadded:: 4.0.0
    """
    return Flip(func)


def flow(*funcs):
    """Creates a function that is the composition of the provided functions,
    where each successive invocation is supplied the return value of the
    previous. For example, composing the functions ``f()``, ``g()``, and
    ``h()`` produces ``h(g(f()))``.

    Args:
        *funcs (function): Function(s) to compose.

    Returns:
        Flow: Function(s) wrapped in a :class:`Flow` context.

    Example:

        >>> mult_5 = lambda x: x * 5
        >>> div_10 = lambda x: x / 10.0
        >>> pow_2 = lambda x: x ** 2
        >>> ops = flow(sum, mult_5, div_10, pow_2)
        >>> ops([1, 2, 3, 4])
        25.0

    .. versionadded:: 2.0.0

    .. versionchanged:: 2.3.1
        Added :func:`pipe` as alias.

    .. versionchanged:: 4.0.0
        Removed alias ``pipe``.
    """
    return Flow(*funcs, from_right=False)


def flow_right(*funcs):
    """This function is like :func:`flow` except that it creates a function
    that invokes the provided functions from right to left. For example,
    composing the functions ``f()``, ``g()``, and ``h()`` produces
    ``f(g(h()))``.

    Args:
        *funcs (function): Function(s) to compose.

    Returns:
        Flow: Function(s) wrapped in a :class:`Flow` context.

    Example:

        >>> mult_5 = lambda x: x * 5
        >>> div_10 = lambda x: x / 10.0
        >>> pow_2 = lambda x: x ** 2
        >>> ops = flow_right(mult_5, div_10, pow_2, sum)
        >>> ops([1, 2, 3, 4])
        50.0

    .. versionadded:: 1.0.0

    .. versionchanged:: 2.0.0
        Added :func:`flow_right` and made :func:`compose` an alias.

    .. versionchanged:: 2.3.1
        Added :func:`pipe_right` as alias.

    .. versionchanged:: 4.0.0
        Removed aliases ``pipe_right`` and ``compose``.
    """
    return Flow(*funcs, from_right=True)


def iterated(func):
    """Creates a function that is composed with itself. Each call to the
    iterated function uses the previous function call's result as input.
    Returned :class:`Iterated` instance can be called with ``(initial, n)``
    where `initial` is the initial value to seed `func` with and `n` is the
    number of times to call `func`.

    Args:
        func (function): Function to iterate.

    Returns:
        Iterated: Function wrapped in a :class:`Iterated` context.

    Example:

        >>> doubler = iterated(lambda x: x * 2)
        >>> doubler(4, 5)
        128
        >>> doubler(3, 9)
        1536

    .. versionadded:: 2.0.0
    """
    return Iterated(func)


def juxtapose(*funcs):
    """Creates a function whose return value is a list of the results of
    calling each `funcs` with the supplied arguments.

    Args:
        *funcs (function): Function(s) to juxtapose.

    Returns:
        Juxtapose: Function wrapped in a :class:`Juxtapose` context.

    Example:

        >>> double = lambda x: x * 2
        >>> triple = lambda x: x * 3
        >>> quadruple = lambda x: x * 4
        >>> juxtapose(double, triple, quadruple)(5)
        [10, 15, 20]


    .. versionadded:: 2.0.0
    """
    return Juxtapose(*funcs)


def negate(func):
    """Creates a function that negates the result of the predicate `func`. The
    `func` function is executed with the arguments of the created function.

    Args:
        func (function): Function to negate execute.

    Returns:
        Negate: Function wrapped in a :class:`Negate` context.

    Example:

        >>> not_is_number = negate(lambda x: isinstance(x, (int, float)))
        >>> not_is_number(1)
        False
        >>> not_is_number('1')
        True

    .. versionadded:: 1.1.0
    """
    return Negate(func)


def once(func):
    """Creates a function that is restricted to execute `func` once. Repeat
    calls to the function will return the value of the first call.

    Args:
        func (function): Function to execute.

    Returns:
        Once: Function wrapped in a :class:`Once` context.

    Example:

        >>> oncer = once(lambda *args: args[0])
        >>> oncer(5)
        5
        >>> oncer(6)
        5

    .. versionadded:: 1.0.0
    """
    return Once(func)


def over_args(func, *transforms):
    """Creates a function that runs each argument through a corresponding
    transform function.

    Args:
        func (function): Function to wrap.
        *transforms (function): Functions to transform arguments, specified as
            individual functions or lists of functions.

    Returns:
        OverArgs: Function wrapped in a :class:`OverArgs` context.

    Example:

        >>> squared = lambda x: x ** 2
        >>> double = lambda x: x * 2
        >>> modder = over_args(lambda x, y: [x, y], squared, double)
        >>> modder(5, 10)
        [25, 20]

    .. versionadded:: 3.3.0

    .. versionchanged:: 4.0.0
        Renamed from ``mod_args`` to ``over_args``.
    """
    return OverArgs(func, *transforms)


def partial(func, *args, **kargs):
    """Creates a function that, when called, invokes `func` with any additional
    partial arguments prepended to those provided to the new function.

    Args:
        func (function): Function to execute.
        *args (optional): Partial arguments to prepend to function call.
        **kargs (optional): Partial keyword arguments to bind to function call.

    Returns:
        Partial: Function wrapped in a :class:`Partial` context.

    Example:

        >>> dropper = partial(lambda array, n: array[n:], [1, 2, 3, 4])
        >>> dropper(2)
        [3, 4]
        >>> dropper(1)
        [2, 3, 4]
        >>> myrest = partial(lambda array, n: array[n:], n=1)
        >>> myrest([1, 2, 3, 4])
        [2, 3, 4]

    .. versionadded:: 1.0.0
    """
    return Partial(func, args, kargs)


def partial_right(func, *args, **kargs):
    """This method is like :func:`partial` except that partial arguments are
    appended to those provided to the new function.

    Args:
        func (function): Function to execute.
        *args (optional): Partial arguments to append to function call.
        **kargs (optional): Partial keyword arguments to bind to function call.

    Returns:
        Partial: Function wrapped in a :class:`Partial` context.

    Example:

        >>> myrest = partial_right(lambda array, n: array[n:], 1)
        >>> myrest([1, 2, 3, 4])
        [2, 3, 4]

    .. versionadded:: 1.0.0
    """
    return Partial(func, args, kargs, from_right=True)


def rearg(func, *indexes):
    """Creates a function that invokes `func` with arguments arranged according
    to the specified indexes where the argument value at the first index is
    provided as the first argument, the argument value at the second index is
    provided as the second argument, and so on.

    Args:
        func (function): Function to rearrange arguments for.
        *indexes (int): The arranged argument indexes.

    Returns:
        Rearg: Function wrapped in a :class:`Rearg` context.

    Example:

        >>> jumble = rearg(lambda *args: args, 1, 2, 3)
        >>> jumble(1, 2, 3)
        (2, 3, 1)
        >>> jumble('a', 'b', 'c', 'd', 'e')
        ('b', 'c', 'd', 'a', 'e')

    .. versionadded:: 3.0.0
    """
    return Rearg(func, *indexes)


def spread(func):
    """Creates a function that invokes `func` with the array of arguments
    provided to the created function.

    Args:
        func (function): Function to spread.

    Returns:
        Spread: Function wrapped in a :class:`Spread` context.

    Example:

        >>> greet = spread(lambda people: 'Hello ' + ', '.join(people) + '!')
        >>> greet(['Mike', 'Don', 'Leo'])
        'Hello Mike, Don, Leo!'

    .. versionadded:: 3.1.0
    """
    return Spread(func)


def throttle(func, wait):
    """Creates a function that, when executed, will only call the `func`
    function at most once per every `wait` milliseconds. Subsequent calls to
    the throttled function will return the result of the last `func` call.

    Args:
        func (function): Function to throttle.
        wait (int): Milliseconds to wait before calling `func` again.

    Returns:
        mixed: Results of last `func` call.

    .. versionadded:: 1.0.0
    """
    return Throttle(func, wait)


def unary(func):
    """Creates a function that accepts up to one argument, ignoring any
    additional arguments.

    Args:
        func (function): Function to cap arguments for.

    Returns:
        Ary: Function wrapped in an :class:`Ary` context.

    Example:

        >>> func = lambda a, b=1, c=0, d=5: (a, b, c, d)
        >>> unary_func = unary(func)
        >>> unary_func(1, 2, 3, 4, 5, 6)
        (1, 1, 0, 5)
        >>> unary_func(1, 2, 3, 4, 5, 6, b=0, c=10, d=20)
        (1, 0, 10, 20)

    .. versionadded:: 4.0.0
    """
    return Ary(func, 1)


def wrap(value, func):
    """Creates a function that provides value to the wrapper function as its
    first argument. Additional arguments provided to the function are appended
    to those provided to the wrapper function.

    Args:
        value (mixed): Value provided as first argument to function call.
        func (function): Function to execute.

    Returns:
        Partial: Function wrapped in a :class:`Partial` context.

    Example:

        >>> wrapper = wrap('hello', lambda *args: args)
        >>> wrapper(1, 2)
        ('hello', 1, 2)

    .. versionadded:: 1.0.0
    """
    return Partial(func, (value,))



# -*- coding: utf-8 -*-

import time

import pytest
import mock

import pydash as _

from .fixtures import parametrize, mocked_sleep


@parametrize('case,expected', [
    ((lambda a, b: a / b, 4, 2), 2)
])
def test_attempt(case, expected):
    assert _.attempt(*case) == expected


@parametrize('case,expected', [
    ((lambda a, b: a / b, 4, 0), ZeroDivisionError)
])
def test_attempt_exception(case, expected):
    assert isinstance(_.attempt(*case), expected)


@parametrize('pairs,case,expected', [
    (([_.matches({'b': 2}), _.constant('matches B')],
      [_.matches({'a': 1}), _.constant('matches A')]),
     {'a': 1, 'b': 2},
     'matches B'),
    (([_.matches({'b': 2}), _.constant('matches B')],
      [_.matches({'a': 1}), _.invert]),
     {'a': 1, 'b': 3},
     {1: 'a', 3: 'b'}),
    (([_.matches({'a': 1}), _.constant('matches A')],
      [_.matches({'b': 2}), _.constant('matches B')]),
     {'a': 1, 'b': 2},
     'matches A')
])
def test_cond(pairs, case, expected):
    func = _.cond(*pairs)
    assert func(case) == expected


@parametrize('case,expected', [
    ([_.matches({'b': 2})], ValueError),
    ([_.matches({'b': 2}), _.matches({'a': 1}), _.constant('matches B')],
     ValueError),
    ([1, 2], ValueError),
    ([[1, 2]], TypeError),
    ([_.matches({'b': 2}), 2], ValueError)
])
def test_cond_exception(case, expected):
    with pytest.raises(expected):
        _.cond(case)


@parametrize('source,case,expected', [
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 0, 'b': -1},
     True),
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 1, 'b': -1},
     False),
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 0, 'b': 1},
     False),
    ({'b': lambda n: n > 1},
     {'b': 2},
     True),
    ({'b': lambda n: n > 1},
     {'b': 0},
     False),
    ([lambda n: n == 0, lambda n: n < 1],
     [0, -1],
     True),
    ([lambda n: n == 0, lambda n: n < 1],
     [1, -1],
     False),
    ([lambda n: n == 0, lambda n: n < 1],
     [0, 1],
     False),
])
def test_conforms(source, case, expected):
    func = _.conforms(source)
    assert func(case) == expected


@parametrize('source,case,expected', [
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 0, 'b': -1},
     True),
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 1, 'b': -1},
     False),
    ({'a': lambda n: n == 0, 'b': lambda n: n < 0},
     {'a': 0, 'b': 1},
     False),
    ({'b': lambda n: n > 1},
     {'b': 2},
     True),
    ({'b': lambda n: n > 1},
     {'b': 0},
     False),
    ([lambda n: n == 0, lambda n: n < 1],
     [0, -1],
     True),
    ([lambda n: n == 0, lambda n: n < 1],
     [1, -1],
     False),
    ([lambda n: n == 0, lambda n: n < 1],
     [0, 1],
     False),
])
def test_conforms_to(source, case, expected):
    assert _.conforms_to(case, source) == expected


@parametrize('case', [
    'foo',
    'bar'
])
def test_constant(case):
    assert _.constant(case)() == case


@parametrize('case,expected', [
    (([1, 10]), 1),
    (([None, 10]), 10),
    (([[1, 2], [3, 4]]), [1, 2]),
    (([None, [3, 4]]), [3, 4])
])
def test_default_to(case, expected):
    assert _.default_to(*case) == expected


@parametrize('case,expected', [
    ((1,), 1),
    ((1, 2), 1),
    ((), None)
])
def test_identity(case, expected):
    assert _.identity(*case) == expected


@parametrize('case,arg,expected', [
    ('name',
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     ['fred', 'barney']),
    ('friends.1',
     [{'name': 'fred', 'age': 40, 'friends': ['barney', 'betty']},
      {'name': 'barney', 'age': 36, 'friends': ['fred', 'wilma']}],
     ['betty', 'wilma']),
    (['friends.1'],
     [{'name': 'fred', 'age': 40, 'friends.1': 'betty'},
      {'name': 'barney', 'age': 36, 'friends.1': 'wilma'}],
     ['betty', 'wilma']),
    ({'name': 'fred'},
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     [True, False]),
    (lambda obj: obj['age'],
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     [40, 36]),
    (None,
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}]),
    (['name', 'fred'],
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     [True, False]),
    (1, [[0, 1], [2, 3], [4, 5]], [1, 3, 5]),
    (('a', 'b.c', ['d', 0, 'f']),
     [{'a': 1, 'b': {'c': 2}, 'd': [{'f': 3}, 4]}],
     [[1, 2, 3]]),
    (('a',),
     [{'a': 1}],
     [[1]]),
])
def test_iteratee(case, arg, expected):
    getter = _.iteratee(case)
    assert _.map_(arg, getter) == expected


@parametrize('case,arg,expected', [
    ({'age': 36}, {'name': 'barney', 'age': 36}, True),
    ({'age': 36}, {'name': 'barney', 'age': 40}, False),
    ({'a': {'b': 2}}, {'a': {'b': 2, 'c': 3}}, True),
])
def test_matches(case, arg, expected):
    assert _.matches(case)(arg) is expected


@parametrize('case,arg,expected', [
    (('a', 1), {'a': 1, 'b': 2}, True),
    (('a', 2), {'a': 1, 'b': 2}, False),
    ((1, 2), [1, 2, 3], True),
    ((1, 3), [1, 2, 3], False),
])
def test_matches_property(case, arg, expected):
    assert _.matches_property(*case)(arg) is expected


@parametrize('case,args,kargs,key', [
    ((lambda a, b: a + b,), (1, 2), {}, '(1, 2){}'),
    ((lambda a, b: a + b,), (1,), {'b': 2}, "(1,){'b': 2}"),
    ((lambda a, b: a + b, lambda a, b: a * b), (1, 2), {}, 2),
    ((lambda a, b: a + b, lambda a, b: a * b), (1,), {'b': 2}, 2),
])
def test_memoize(case, args, kargs, key):
    memoized = _.memoize(*case)
    expected = case[0](*args, **kargs)
    assert memoized(*args, **kargs) == expected
    assert memoized.cache
    assert memoized.cache[key] == expected


@parametrize('case,args,kargs,expected', [
    (('a.b',), ({'a': {'b': lambda x, y: x + y}}, 1, 2), {}, 3),
    (('a.b',), ({'a': {'b': lambda x, y: x + y}}, 1,), {'y': 2}, 3),
    (('a.b', 1), ({'a': {'b': lambda x, y: x + y}}, 2,), {}, 3),
])
def test_method(case, args, kargs, expected):
    assert _.method(*case)(*args, **kargs) == expected


@parametrize('case,args,kargs,expected', [
    (({'a': {'b': lambda x, y: x + y}},), ('a.b', 1, 2), {}, 3),
    (({'a': {'b': lambda x, y: x + y}},), ('a.b', 1,), {'y': 2}, 3),
])
def test_method_of(case, args, kargs, expected):
    assert _.method_of(*case)(*args, **kargs) == expected


@parametrize('case,expected', [
    ((), None),
    ((1, 2, 3), None)
])
def test_noop(case, expected):
    assert _.noop(*case) == expected


@parametrize('args,pos,expected', [
    ([11, 22, 33, 44], 0, 11),
    ([11, 22, 33, 44], -1, 44),
    ([11, 22, 33, 44], -4, 11),
    ([11, 22, 33, 44], -5, None),
    ([11, 22, 33, 44], 4, None),
    ([11, 22, 33], '1', 22),
    ([11, 22, 33], 'xyz', 11),
    ([11, 22, 33], 1.45, 33),
    ([11, 22, 33], 1.51, 33)
])
def test_nth_arg(args, pos, expected):
    func = _.nth_arg(pos)
    assert func(*args) == expected


def test_now():
    present = int(time.time() * 1000)
    # Add some leeway when comparing time.
    assert (present - 1) <= _.now() <= (present + 1)


@parametrize('funcs,data,expected', [
    ([max, min], [1, 2, 3, 4], [4, 1])
])
def test_over(funcs, data, expected):
    assert _.over(funcs)(*data) == expected


@parametrize('funcs,data,expected', [
    ([lambda x: x is not None, bool], [1], True),
    ([lambda x: x is None, bool], [1], False)
])
def test_over_every(funcs, data, expected):
    assert _.over_every(funcs)(*data) == expected


@parametrize('funcs,data,expected', [
    ([lambda x: x is not None, bool], [1], True),
    ([lambda x: x is None, bool], [1], True),
    ([lambda x: x is False, lambda y: y == 2], [True], False)
])
def test_over_some(funcs, data, expected):
    assert _.over_some(funcs)(*data) == expected


@parametrize('case,arg,expected', [
    ('name',
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     ['fred', 'barney']),
    ('name',
     [{'name': 'fred', 'age': 40},
      {'name': 'barney', 'age': 36}],
     ['fred', 'barney']),
    ('spouse.name',
     [{'name': 'fred', 'age': 40, 'spouse': {'name': 'wilma'}},
      {'name': 'barney', 'age': 36, 'spouse': {'name': 'betty'}}],
     ['wilma', 'betty'])
])
def test_property_(case, arg, expected):
    assert _.map_(arg, _.property_(case)) == expected


@parametrize('case,arg,expected', [
    ({'name': 'fred', 'age': 40}, ['name', 'age'], ['fred', 40]),
])
def test_property_of(case, arg, expected):
    assert _.map_(arg, _.property_of(case)) == expected


@parametrize('case,minimum,maximum', [
    ((), 0, 1),
    ((25,), 0, 25),
    ((5, 10), 5, 10)
])
def test_random(case, minimum, maximum):
    for x in range(50):
        rnd = _.random(*case)
        assert isinstance(rnd, int)
        assert minimum <= rnd <= maximum


@parametrize('case,floating,minimum,maximum', [
    ((), True, 0, 1),
    ((25,), True, 0, 25),
    ((5, 10), True, 5, 10),
    ((5.0, 10), False, 5, 10),
    ((5, 10.0), False, 5, 10),
    ((5.0, 10.0), False, 5, 10),
    ((5.0, 10.0), True, 5, 10),
])
def test_random_float(case, floating, minimum, maximum):
    for x in range(50):
        rnd = _.random(*case, floating=floating)
        assert isinstance(rnd, float)
        assert minimum <= rnd <= maximum


@parametrize('case,expected', [
    ((10,), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
    ((1, 11), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
    ((11, 1), [11, 10, 9, 8, 7, 6, 5, 4, 3, 2]),
    ((11, 1, -2), [11, 9, 7, 5, 3]),
    ((11, 1, 2), []),
    ((0, 30, 5), [0, 5, 10, 15, 20, 25]),
    ((0, -10, -1), [0, -1, -2, -3, -4, -5, -6, -7, -8, -9]),
    ((0,), []),
    ((), []),
])
def test_range_(case, expected):
    assert list(_.range_(*case)) == expected


@parametrize('case,expected', [
    ((4,), [3, 2, 1, 0]),
    ((-4,), [-3, -2, -1, 0]),
    ((1, 5), [4, 3, 2, 1]),
    ((0, 20, 5), [15, 10, 5, 0]),
    ((0, -20, -5), [-15, -10, -5, -0]),
    ((0, -4, -1), [-3, -2, -1, 0]),
    ((1, 10), [9, 8, 7, 6, 5, 4, 3, 2, 1]),
    ((-1, 10), [9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1]),
    ((-1, -10), [-9, -8, -7, -6, -5, -4, -3, -2, -1]),
    ((1, 10, 2), [9, 7, 5, 3, 1]),
    ((1, 10, -2), []),
    ((-1, 10, 2), [9, 7, 5, 3, 1, -1]),
    ((-1, 10, -2), []),
    ((-1, -10, 2), []),
    ((-1, -10, -2), [-9, -7, -5, -3, -1]),
    ((10, 1), [2, 3, 4, 5, 6, 7, 8, 9, 10]),
    ((-10, 1), [0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10]),
    ((-10, -1), [-2, -3, -4, -5, -6, -7, -8, -9, -10]),
    ((10, 1, 2), []),
    ((10, 1, -2), [2, 4, 6, 8, 10]),
    ((-10, 1, 2), [0, -2, -4, -6, -8, -10]),
    ((-10, 1, -2), []),
    ((-10, -1, 2), [-2, -4, -6, -8, -10]),
    ((-10, -1, -2), []),
    ((1, 4, 0), [1, 1, 1]),
    ((0,), []),
    ((10,), [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]),
    ((1, 5), [4, 3, 2, 1]),
    ((0, 20, 5), [15, 10, 5, 0]),
    ((0, -4, -1), [-3, -2, -1, 0]),
    ((0,), []),
    ((), []),
])
def test_range_right(case, expected):
    assert list(_.range_right(*case)) == expected


@parametrize('case,expected', [
    (({'cheese': 'crumpets', 'stuff': lambda: 'nonsense'}, 'cheese'),
     'crumpets'),
    (({'cheese': 'crumpets', 'stuff': lambda: 'nonsense'}, 'stuff'),
     'nonsense'),
    (({'cheese': 'crumpets', 'stuff': lambda: 'nonsense'}, 'foo'),
     None),
    ((False, 'foo'), None),
    ((False, 'foo', 'default'), 'default'),
])
def test_result(case, expected):
    assert _.result(*case) == expected


@parametrize('case,delay_count,delay_times', [
    ({}, 2, [0.5, 1.0]),
    ({'attempts': 1}, 0, []),
    ({'attempts': 3, 'delay': 0.5, 'scale': 2.0}, 2, [0.5, 1.0]),
    ({'attempts': 5, 'delay': 1.5, 'scale': 2.5}, 4,
     [1.5, 3.75, 9.375, 23.4375]),
    ({'attempts': 5, 'delay': 1.5, 'max_delay': 8.0, 'scale': 2.5}, 4,
     [1.5, 3.75, 8.0, 8.0]),
])
def test_retry(mocked_sleep, case, delay_count, delay_times):
    @_.retry(**case)
    def func():
        raise Exception()

    with pytest.raises(Exception):
        func()

    assert delay_count == mocked_sleep.call_count

    delay_calls = [mock.call(time) for time in delay_times]
    assert delay_calls == mocked_sleep.call_args_list


@parametrize('case,delay_count', [
    ({'attempts': 3}, 0),
    ({'attempts': 3}, 1),
    ({'attempts': 3}, 2),
    ({'attempts': 5}, 3)
])
def test_retry_success(mocked_sleep, case, delay_count):
    counter = {True: 0}

    @_.retry(**case)
    def func():
        if counter[True] != delay_count:
            counter[True] += 1
            raise Exception()
        return True

    result = func()

    assert result is True
    assert counter[True] == delay_count
    assert delay_count == mocked_sleep.call_count


@parametrize('case,unexpected_delay_times', [
    ({'jitter': 5, 'delay': 2, 'scale': 1, 'attempts': 5},
     [2, 2, 2, 2]),
    ({'jitter': 10, 'delay': 3, 'scale': 1.5, 'attempts': 5},
     [3, 4.5, 6.75, 10.125]),
    ({'jitter': 1.0, 'delay': 3, 'scale': 1.5, 'attempts': 5},
     [3, 4.5, 6.75, 10.125]),
])
def test_retry_jitter(mocked_sleep, case, unexpected_delay_times):
    @_.retry(**case)
    def func():
        raise Exception()

    with pytest.raises(Exception):
        func()

    unexpected_delay_calls = [mock.call(time)
                              for time in unexpected_delay_times]

    assert len(unexpected_delay_calls) == mocked_sleep.call_count
    assert unexpected_delay_calls != mocked_sleep.call_args_list


@parametrize('case,raise_exc,delay_count', [
    ({'attempts': 1, 'exceptions': (RuntimeError,)}, RuntimeError, 0),
    ({'attempts': 2, 'exceptions': (RuntimeError,)}, RuntimeError, 1),
    ({'attempts': 2, 'exceptions': (RuntimeError,)}, Exception, 0),
])
def test_retry_exceptions(mocked_sleep, case, raise_exc, delay_count):
    @_.retry(**case)
    def func():
        raise raise_exc()

    with pytest.raises(raise_exc):
        func()

    assert delay_count == mocked_sleep.call_count


def test_retry_on_exception(mocked_sleep):
    attempts = 5
    error_count = {True: 0}

    def on_exception(exc):
        error_count[True] += 1

    @_.retry(attempts=attempts, on_exception=on_exception)
    def func():
        raise Exception()

    with pytest.raises(Exception):
        func()

    assert error_count[True] == attempts


@parametrize('case,exception', [
    ({'attempts': 0}, ValueError),
    ({'attempts': '1'}, ValueError),
    ({'delay': -1}, ValueError),
    ({'delay': '1'}, ValueError),
    ({'max_delay': -1}, ValueError),
    ({'max_delay': '1'}, ValueError),
    ({'scale': 0}, ValueError),
    ({'scale': '1'}, ValueError),
    ({'jitter': -1}, ValueError),
    ({'jitter': '1'}, ValueError),
    ({'jitter': (1,)}, ValueError),
    ({'jitter': ('1', '2')}, ValueError),
    ({'exceptions': (1, 2)}, TypeError),
    ({'exceptions': 1}, TypeError),
    ({'exceptions': (Exception, 2)}, TypeError),
    ({'on_exception': 5}, TypeError),
])
def test_retry_invalid_args(case, exception):
    with pytest.raises(exception):
        _.retry(**case)


@parametrize('case,expected', [
    (_.times(2, _.stub_list), [[], []]),
    (_.stub_list(), [])
])
def test_stub_list(case, expected):
    assert case == expected


@parametrize('case,expected', [
    (_.times(2, _.stub_dict), [{}, {}]),
    (_.stub_dict(), {})
])
def test_stub_dict(case, expected):
    assert case == expected


@parametrize('case,expected', [
    (_.times(2, _.stub_false), [False, False]),
    (_.stub_false(), False)
])
def test_stub_false(case, expected):
    assert case == expected


@parametrize('case,expected', [
    (_.times(2, _.stub_string), ['', '']),
    (_.stub_string(), '')
])
def test_stub_string(case, expected):
    assert case == expected


@parametrize('case,expected', [
    (_.times(2, _.stub_true), [True, True]),
    (_.stub_true(), True)
])
def test_stub_true(case, expected):
    assert case == expected


@parametrize('case,expected', [
    ((5, lambda i: i * i), [0, 1, 4, 9, 16])
])
def test_times(case, expected):
    assert _.times(*case) == expected


@parametrize('case,expected', [
    ('a.b.c', ['a', 'b', 'c']),
    ('a[0].b.c', ['a', 0, 'b', 'c']),
    ('a[0][1][2].b.c', ['a', 0, 1, 2, 'b', 'c'])
])
def test_to_path(case, expected):
    assert _.to_path(case) == expected


def test_unique_id_setup():
    _.utilities.ID_COUNTER = 0


@parametrize('case,expected', [
    ((), '1'),
    (('foo',), 'foo2')
])
def test_unique_id(case, expected):
    assert _.unique_id(*case) == expected



# -*- coding: utf-8 -*-
"""Generic utility methods not part of main API.
"""

from __future__ import absolute_import

from collections import Iterable
from functools import wraps
from operator import attrgetter, itemgetter
import warnings

import pydash as pyd
from ._compat import PY2, iteritems, getfullargspec, string_types


class _NoValue(object):
    """Represents an unset value. Used to differeniate between an explicit
    ``None`` and an unset value.
    """
    pass


#: Singleton object that differeniates between an explicit ``None`` value and
#: an unset value.
NoValue = _NoValue()


def callit(iteratee, *args, **kargs):
    """Inspect argspec of `iteratee` function and only pass the supported
    arguments when calling it.
    """
    maxargs = len(args)
    argcount = (kargs['argcount'] if 'argcount' in kargs
                else getargcount(iteratee, maxargs))
    argstop = min([maxargs, argcount])

    return iteratee(*args[:argstop])


def getargcount(iteratee, maxargs):
    """Return argument count of iteratee function."""
    if hasattr(iteratee, '_argcount'):
        # Optimization feature where argcount of iteratee is known and properly
        # set by initator.
        return iteratee._argcount

    if isinstance(iteratee, type) or pyd.is_builtin(iteratee):
        # Only pass single argument to type iteratees or builtins.
        argcount = 1
    else:
        argcount = 1

        try:
            argcount = _getfullargspec(iteratee, maxargs)
        except TypeError:  # pragma: no cover
            # PY2: Python2.7 throws a TypeError on classes that have __call__()
            # defined but Python3 doesn't. So if we fail with TypeError here,
            # try iteratee as iteratee.__call__.
            if PY2 and hasattr(iteratee, '__call__'):
                try:
                    argcount = _getfullargspec(iteratee.__call__, maxargs)
                except TypeError:
                    pass

    return argcount


def _getfullargspec(iteratee, maxargs):
    argspec = getfullargspec(iteratee)

    if argspec and not argspec.varargs:
        # Use inspected arg count.
        argcount = len(argspec.args)
    else:
        # Assume all args are handleable
        argcount = maxargs

    return argcount


def iteriteratee(obj, iteratee=None, reverse=False):
    """Return iterative iteratee based on collection type."""
    cbk = pyd.iteratee(iteratee)
    items = iterator(obj)

    if reverse:
        items = reversed(tuple(items))

    # Precompute argcount to avoid repeated calculations during iteratee loop.
    argcount = getargcount(cbk, maxargs=3)

    for key, item in items:
        yield (callit(cbk, item, key, obj, argcount=argcount),
               item,
               key,
               obj)


def iterator(obj):
    """Return iterative based on object type."""
    if isinstance(obj, dict):
        return iteritems(obj)
    elif hasattr(obj, 'iteritems'):
        return obj.iteritems()
    elif hasattr(obj, 'items'):
        return iter(obj.items())
    elif isinstance(obj, Iterable):
        return enumerate(obj)
    else:
        return iteritems(getattr(obj, '__dict__', {}))


def base_get(obj, key, default=NoValue):
    """Safely get an item by `key` from a sequence or mapping object when
    `default` provided.

    Args:
        obj (list|dict): Sequence or mapping to retrieve item from.
        key (mixed): Key or index identifying which item to retrieve.

    Keyword Args:
        use_default (bool, optional): Whether to use `default` value when
            `key` doesn't exist in `obj`.
        default (mixed, optional): Default value to return if `key` not
            found in `obj`.

    Returns:
        mixed: `obj[key]`, `obj.key`, or `default`.

    Raises:
        KeyError: If `obj` is missing key, index, or attribute and no default
            value provided.
    """
    # Build list of getters to try to retrieve key value from obj.
    getters = [itemgetter(key)]

    try:
        # Only add list index getter if key can be cast as integer.
        getters.append(itemgetter(int(key)))
    except Exception:
        pass

    if not isinstance(obj, (dict, list)):
        # Don't add attrgetter for dict/list objects since we don't want class
        # methods/attributes returned for them.
        try:
            # Only add attribute getter if key is string.
            getters.append(attrgetter(key))
        except Exception:  # pragma: no cover
            pass

    for getter in getters:
        try:
            ret = getter(obj)
            break
        except Exception:  # pragma: no cover
            pass
    else:
        # The for-loop didn't break which means we weren't able to find key.
        if default is NoValue:
            # Raise if there's no default provided.
            raise KeyError('Object "{0}" does not have key "{1}"'
                           .format(repr(obj), key))
        ret = default

    return ret


def base_set(obj, key, value, allow_override=True):
    """Set an object's `key` to `value`. If `obj` is a ``list`` and the
    `key` is the next available index position, append to list; otherwise, pad
    the list of ``None`` and then append to the list.

    Args:
        obj (list|dict): Object to assign value to.
        key (mixed): Key or index to assign to.
        value (mixed): Value to assign.
    """
    if isinstance(obj, dict):
        if allow_override or key not in obj:
            obj[key] = value
    elif isinstance(obj, list):
        key = int(key)

        if key < len(obj):
            if allow_override:
                obj[key] = value
        else:
            if key > len(obj):
                # Pad list object with None values up to the index key so we
                # can append the value into the key index.
                obj[:] = (obj + [None] * key)[:key]
            obj.append(value)

    return obj


def parse_iteratee(iteratee_keyword, *args, **kargs):
    """Try to find iteratee function passed in either as a keyword argument or
    as the last positional argument in `args`.
    """
    iteratee = kargs.get(iteratee_keyword)
    last_arg = args[-1]

    if (iteratee is None and
            (callable(last_arg) or
             isinstance(last_arg, string_types) or
             isinstance(last_arg, dict) or
             last_arg is None)):
        iteratee = last_arg
        args = args[:-1]

    return (iteratee, args)


class iterator_with_default(object):
    """A wrapper around an iterator object that provides a default."""
    def __init__(self, collection, default):
        self.iter = iter(collection)
        self.default = default

    def __iter__(self):
        return self

    def next_default(self):
        ret = self.default
        self.default = NoValue
        return ret

    def __next__(self):
        ret = next(self.iter, self.next_default())
        if ret is NoValue:
            raise StopIteration
        return ret

    next = __next__


def deprecated(func):  # pragma: no cover
    """This is a decorator which can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used.
    """
    @wraps(func)
    def wrapper(*args, **kargs):
        warnings.warn('Call to deprecated function {0}.'.format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=3)
        return func(*args, **kargs)

    return wrapper
